from typing import Optional


class TrackerDebug:
    def __init__(self):
        self.is_debug = False
        self.values = {}

    def store(self, name: str, value: any):
        if not self.is_debug:
            return

        if name not in self.values:
            self.values[name] = []

        self.values[name].append(value)

    def get(self, name: str):
        return self.values.get(name, None)


_internal: Optional[TrackerDebug] = None


def tracker_debug_singleton() -> TrackerDebug:
    global _internal
    if _internal is None:
        _internal = TrackerDebug()

    return _internal



from typing import List, Callable, overload, Union, Tuple

from labml.internal.configs.base import Configs as _Configs
from labml.internal.configs.config_item import ConfigItem
from labml.utils.errors import ConfigsError


class BaseConfigs(_Configs):
    r"""
    You should sub-class this class to create your own configurations
    """
    pass


def _get_config_class(name: any):
    if isinstance(name, ConfigItem):
        return name.configs_class
    elif isinstance(name, list) and len(name) > 0 and isinstance(name[0], ConfigItem):
        return name[0].configs_class
    else:
        return None


@overload
def option(name: Union[any, List[any]]):
    ...


@overload
def option(name: Union[any, List[any]], option_name: str):
    ...


@overload
def option(name: Union[any, List[any]], pass_params: List[any]):
    ...


@overload
def option(name: Union[any, List[any]], option_name: str, pass_params: List[any]):
    ...


def option(name: Union[any, List[any]], *args: any):
    r"""
    Use this as a decorator to register configuration options.

    This has multiple overloads

    .. function:: option(config_item: Union[any, List[any]])
        :noindex:

    .. function:: option(config_item: Union[any, List[any]], option_name: str)
        :noindex:

    .. function:: option(config_item: Union[any, List[any]], pass_params: List[any])
        :noindex:

    .. function:: option(config_item: Union[any, List[any]], option_name: str, pass_params: List[any])
        :noindex:

    Arguments:
        name: the configuration item or a list of items.
            If it is a list of items the function should return
            tuple.

        option_name (str, optional): name of the option.
            If not provided it will be derived from the
            function name.

        pass_params (list, optional): list of params to be passed.
            If not provided the configs object is passed.
            If provided the corresponding calculated configuration items
            will be passed to the function
    """
    config_class = _get_config_class(name)
    if config_class is None:
        raise ConfigsError('You need to pass config items to option')

    option_name = None
    pass_params = None
    for arg in args:
        if isinstance(arg, str):
            option_name = arg
        elif isinstance(arg, list):
            pass_params = arg

    return config_class.calc(name, option_name, pass_params)


@overload
def calculate(name: Union[any, List[any]], func: Callable):
    ...


@overload
def calculate(name: Union[any, List[any]], option_name: str, func: Callable):
    ...


@overload
def calculate(name: Union[any, List[any]],
              pass_params: List[any], func: Callable):
    ...


@overload
def calculate(name: Union[any, List[any]], option_name: str,
              pass_params: List[any], func: Callable):
    ...


def calculate(name: any, *args: any):
    r"""
    Use this to register configuration options.

    This has multiple overloads

    .. function:: calculate(name: Union[any, List[any]], func: Callable)
        :noindex:

    .. function:: calculate(name: Union[any, List[any]], option_name: str, func: Callable)
        :noindex:

    .. function:: calculate(name: Union[any, List[any]], pass_params: List[any], func: Callable)
        :noindex:

    .. function:: calculate(name: Union[any, List[any]], option_name: str, pass_params: List[any], func: Callable)
        :noindex:

    Arguments:
        name: the configuration item or a list of items.
            If it is a list of items the function should return
            tuple.

        func: the function to calculate the configuration

        option_name (str, optional): name of the option.
            If not provided it will be derived from the
            function name.

        pass_params (list, optional): list of params to be passed.
            If not provided the configs object is passed.
            If provided the corresponding calculated configuration items
            will be passed to the function
    """
    config_class = _get_config_class(name)
    if config_class is None:
        raise ConfigsError('You need to pass config items to calculate')

    option_name = None
    pass_params = None
    func = None
    for arg in args:
        if isinstance(arg, str):
            option_name = arg
        elif isinstance(arg, list):
            pass_params = arg
        elif type(arg) == type:
            func = arg
        else:
            func = arg

    if func is None:
        raise ConfigsError('You need to pass the function that calculates the configs')

    return config_class.calc_wrap(func, name, option_name, pass_params)


def hyperparams(*args: any, is_hyperparam=True):
    r"""
    Identifies configuration as (or not) hyper-parameters

    Arguments:
        *args: list of configurations
        is_hyperparam (bool, optional): whether the provided configuration
            items are hyper-parameters. Defaults to ``True``.
    """

    for arg in args:
        config_class = _get_config_class(arg)
        if config_class is None:
            raise ConfigsError('You need to pass config items to set hyperparams')
        config_class.set_hyperparams(arg, is_hyperparam=is_hyperparam)


def meta_config(*args: any, is_meta=True):
    r"""
    Identifies configuration as meta parameter

    Arguments:
        *args: list of configurations
        is_meta (bool, optional): whether the provided configuration
            items are meta. Defaults to ``True``.
    """

    for arg in args:
        config_class = _get_config_class(arg)
        if config_class is None:
            raise ConfigsError('You need to pass config items to set hyperparams')
        config_class.set_meta(arg, is_meta=is_meta)


def aggregate(name: any, option_name: str, *args: Tuple[any, any]):
    r"""
    Aggregate configs

    Arguments:
        name: name of the aggregate
        option_name: aggregate option name
        *args: list of configs to be aggregated
    """

    config_class = _get_config_class(name)
    if config_class is None:
        raise ConfigsError('You need to pass config item to aggregate')

    config_class.aggregate(name, option_name, *args)



from typing import Dict, overload, Optional

from labml.internal.tracker import tracker_singleton as _internal
from labml.internal.track_debug import tracker_debug_singleton as _tracker_debug_singleton


@overload
def debug(is_debug: bool):
    ...


@overload
def debug(name: str):
    ...


@overload
def debug(name: str, value: any):
    ...


def debug(*args):
    if len(args) == 1:
        if isinstance(args[0], bool):
            _tracker_debug_singleton().is_debug = args[0]
        else:
            return _tracker_debug_singleton().get(args[0])
    else:
        _tracker_debug_singleton().store(args[0], args[1])


def set_global_step(global_step: Optional[int]):
    _internal().set_global_step(global_step)


def add_global_step(increment_global_step: int = 1):
    _internal().add_global_step(int(increment_global_step))


def get_global_step() -> int:
    return _internal().global_step


def set_queue(name: str, queue_size: int = 10, is_print: bool = False):
    from labml.internal.tracker.indicators.numeric import Queue
    _internal().add_indicator(Queue(name, queue_size, is_print))


def set_histogram(name: str, is_print: bool = False):
    from labml.internal.tracker.indicators.numeric import Histogram
    _internal().add_indicator(Histogram(name, is_print))


def set_scalar(name: str, is_print: bool = False):
    from labml.internal.tracker.indicators.numeric import Scalar
    _internal().add_indicator(Scalar(name, is_print))


def set_indexed_scalar(name: str):
    from labml.internal.tracker.indicators.indexed import IndexedScalar
    _internal().add_indicator(IndexedScalar(name))


def set_image(name: str, is_print: bool = False, density: Optional[float] = None):
    from labml.internal.tracker.indicators.artifacts import Image
    _internal().add_indicator(Image(name, is_print, density))


def set_text(name: str, is_print: bool = False):
    from labml.internal.tracker.indicators.artifacts import Text
    _internal().add_indicator(Text(name, is_print))


def set_tensor(name: str, is_once: bool = False):
    from labml.internal.tracker.indicators.artifacts import Tensor
    _internal().add_indicator(Tensor(name, is_once=is_once))


def set_indexed_text(name: str, title: Optional[str] = None, is_print: bool = False):
    from labml.internal.tracker.indicators.artifacts import IndexedText
    _internal().add_indicator(IndexedText(name, title, is_print))


def _add_dict(values: Dict[str, any]):
    for k, v in values.items():
        _internal().store(k, v)


@overload
def add(values: Dict[str, any]):
    ...


@overload
def add(name: str, value: any):
    ...


@overload
def add(**kwargs: any):
    ...


def add(*args, **kwargs):
    """
    This has multiple overloads

    .. function:: add(values: Dict[str, any])
        :noindex:

    .. function:: add(name: str, value: any)
        :noindex:

    .. function:: add(**kwargs: any)
        :noindex:
    """
    if len(args) > 2:
        raise TypeError('tracker.add should be called as add(name, value), add(dictionary) or add(k=v,k2=v2...)')

    if len(args) == 0:
        _add_dict(kwargs)
    elif len(args) == 1:
        if kwargs:
            raise TypeError('tracker.add should be called as add(name, value), add(dictionary) or add(k=v,k2=v2...)')
        if not isinstance(args[0], dict):
            raise TypeError('tracker.add should be called as add(name, value), add(dictionary) or add(k=v,k2=v2...)')
        _add_dict(args[0])
    elif len(args) == 2:
        if kwargs:
            raise TypeError('tracker.add should be called as add(name, value), add(dictionary) or add(k=v,k2=v2...)')
        if not isinstance(args[0], str):
            raise TypeError('tracker.add should be called as add(name, value), add(dictionary) or add(k=v,k2=v2...)')
        _internal().store(args[0], args[1])


@overload
def save():
    ...


@overload
def save(global_step: int):
    ...


@overload
def save(values: Dict[str, any]):
    ...


@overload
def save(name: str, value: any):
    ...


@overload
def save(**kwargs: any):
    ...


@overload
def save(global_step: int, values: Dict[str, any]):
    ...


@overload
def save(global_step: int, name: str, value: any):
    ...


@overload
def save(global_step: int, **kwargs: any):
    ...


def save(*args, **kwargs):
    r"""
    This has multiple overloads

    .. function:: save()
        :noindex:

    .. function:: save(global_step: int)
        :noindex:

    .. function:: save(values: Dict[str, any])
        :noindex:

    .. function:: save(name: str, value: any)
        :noindex:

    .. function:: save(**kwargs: any)
        :noindex:

    .. function:: save(global_step: int, values: Dict[str, any])
        :noindex:

    .. function:: save(global_step: int, name: str, value: any)
        :noindex:

    .. function:: save(global_step: int, **kwargs: any)
        :noindex:
    """
    if len(args) > 0 and type(args[0]) == int:
        _internal().set_global_step(args[0])
        args = args[1:]

    if len(args) > 0 or len(kwargs.keys()) > 0:
        add(*args, **kwargs)

    _internal().write()


def new_line():
    r"""
    Prints a new line.

    Equivalent to ``logger.log``, but handles distributed training where only the rank=0
    process is tracking data.
    """
    _internal().new_line()

def namespace(name: str):
    r"""
    Set a namespace for tracking
    """
    return _internal().namespace(name)


def reset():
    r"""
    Reset indicators, for a new experiment
    """
    _internal().reset_store()



import sqlite3
import time
from pathlib import PurePath, Path
from typing import Dict, Optional

import numpy as np
from ..indicators import Indicator
from ..indicators.artifacts import Tensor
from ..indicators.indexed import IndexedIndicator
from ..indicators.numeric import NumericIndicator

from . import Writer as WriteBase


class Writer(WriteBase):
    conn: Optional[sqlite3.Connection]

    def __init__(self, sqlite_path: PurePath, artifacts_path: PurePath):
        super().__init__()

        self.sqlite_path = sqlite_path
        self.artifacts_path = artifacts_path
        self.conn = None
        self.scalars_cache = []
        self.indexed_scalars_cache = []
        self.last_committed = time.time()

    def __connect(self):
        if self.conn is not None:
            return

        self.conn = sqlite3.connect(str(self.sqlite_path))

        try:
            self.conn.execute(f"CREATE TABLE scalars "
                              f"(indicator text, step integer, value real)")
            self.conn.execute(f"CREATE TABLE indexed_scalars "
                              f"(indicator text, step integer, idx integer, value real)")
            self.conn.execute(f"CREATE TABLE tensors "
                              f"(indicator text, step integer, filename text)")

        except sqlite3.OperationalError:
            print('Scalar table exists')

    @staticmethod
    def _parse_key(key: str):
        return key
        # if we name tables
        # return key.replace('.', '_')

    def _write_indicator(self, global_step: int, indicator: Indicator):
        if indicator.is_empty():
            return

        if isinstance(indicator, NumericIndicator):
            value = indicator.get_mean()
            key = self._parse_key(indicator.mean_key)
            self.conn.execute(
                f"INSERT INTO scalars VALUES (?, ?, ?)",
                (key, global_step, value))

        if isinstance(indicator, IndexedIndicator):
            idx, value = indicator.get_index_mean()
            key = self._parse_key(indicator.mean_key)
            data = [(key, global_step, i, v) for i, v in zip(idx, value)]
            self.conn.executemany(
                f"INSERT INTO indexed_scalars VALUES (?, ?, ?, ?)",
                data)

        if isinstance(indicator, Tensor):
            key = self._parse_key(indicator.name)
            for k in indicator.keys():
                tensor = indicator.get_value(k)
                if not indicator.is_once:
                    filename = f'{key}_{global_step}_{k}.npy'
                    self.conn.execute(
                        f"INSERT INTO tensors VALUES (?, ?, ?)",
                        (key, global_step, filename))
                else:
                    filename = f'{key}_{k}.npy'
                self.conn.execute(
                    f"INSERT INTO tensors VALUES (?, ?, ?)",
                    (key, -1, filename))

                artifacts_folder = Path(self.artifacts_path)
                if not artifacts_folder.exists():
                    artifacts_folder.mkdir(parents=True)

                np.save(str(self.artifacts_path / filename), tensor)

    def write(self, *,
              global_step: int,
              indicators: Dict[str, Indicator]):
        self.__connect()

        for ind in indicators.values():
            self._write_indicator(global_step, ind)

        t = time.time()
        if t - self.last_committed > 0.1:
            self.last_committed = t
            self.flush()

    def flush(self):
        if self.conn is not None:
            self.conn.commit()



import typing
from typing import Optional, Iterable, TYPE_CHECKING

from .sections import Section

if TYPE_CHECKING:
    from ..monitor import Monitor


class Iterator:
    r"""
        Note:
            You should use :meth:`labml.logger.iterate`
            or :meth:`labml.monit.enum` to create iterators
            and enumerators.
    """

    _section: Optional[Section]

    def __init__(self, *,
                 logger: 'Monitor',
                 name: str,
                 iterable: typing.Union[Iterable, typing.Sized, int],
                 is_silent: bool,
                 is_timed: bool,
                 total_steps: Optional[int],
                 is_children_silent: bool,
                 is_enumerate: bool,
                 section: Optional[Section]):
        if is_enumerate:
            total_steps = len(iterable)
            iterable = enumerate(iterable)
        if isinstance(iterable, int):
            total_steps = iterable
            iterable = range(total_steps)
        if total_steps is None:
            sized: typing.Sized = iterable
            total_steps = len(sized)

        self.is_children_silent = is_children_silent
        self._logger = logger
        self._name = name
        self._iterable: Iterable = iterable
        self._iterator = Optional[typing.Iterator]
        self._total_steps = total_steps
        self._section = section
        self._is_silent = is_silent
        self._is_timed = is_timed
        self._counter = -1
        self.is_section_specified = section is not None

    def get_estimated_time(self) -> float:
        if self._section:
            return self._section.get_estimated_time()
        else:
            return 0

    def __iter__(self):
        if not self.is_section_specified:
            self._section = self._logger.section(
                self._name,
                is_silent=self._is_silent,
                is_timed=self._is_timed,
                is_partial=False,
                total_steps=self._total_steps,
                is_children_silent=self.is_children_silent,
                is_new_line=True)
        self._iterator = iter(self._iterable)
        if not self.is_section_specified:
            self._section.__enter__()

        return self

    def __next__(self):
        try:
            self._counter += 1
            self._logger.progress(self._counter)
            next_value = next(self._iterator)
        except StopIteration as e:
            if not self.is_section_specified:
                self._section.__exit__(None, None, None)
            raise e

        return next_value



from typing import List

import numpy as np
import tensorflow as tf


class MatPlotLibTensorBoardAnalytics:
    @staticmethod
    def render_matrix(matrix, ax, color):
        from matplotlib.patches import Rectangle
        from matplotlib.collections import PatchCollection

        rows, cols = matrix.shape
        x_ticks = [matrix[0, i] for i in range(1, cols)]
        y_ticks = [matrix[i, 0] for i in range(1, rows)]

        max_density = 0
        for y in range(rows - 2):
            for x in range(cols - 2):
                area = (x_ticks[x + 1] - x_ticks[x]) * (y_ticks[y + 1] - y_ticks[y])
                density = matrix[y + 1, x + 1] / area
                if max_density < density:
                    max_density = density

        boxes = []
        for y in range(rows - 2):
            for x in range(cols - 2):
                width = x_ticks[x + 1] - x_ticks[x]
                height = y_ticks[y + 1] - y_ticks[y]
                area = width * height
                density = matrix[y + 1, x + 1] / area
                density = max(density, 1)
                # alpha = np.log(density) / np.log(max_density)
                alpha = density / max_density
                rect = Rectangle((x_ticks[x], y_ticks[y]), width, height,
                                 alpha=alpha, color=color)
                boxes.append(rect)

        pc = PatchCollection(boxes, match_original=True)

        ax.add_collection(pc)

        # Plot something
        _ = ax.errorbar([], [], xerr=[], yerr=[],
                        fmt='None', ecolor='k')

        return x_ticks[0], x_ticks[-1], y_ticks[0], y_ticks[-1]

    def render_tensors(self, tensors: List[any], axes: np.ndarray, color):
        assert len(axes.shape) == 2
        assert axes.shape[0] * axes.shape[1] == len(tensors)
        x_min = x_max = y_min = y_max = 0

        for i in range(axes.shape[0]):
            for j in range(axes.shape[1]):
                idx = i * axes.shape[1] + j
                axes[i, j].set_title(f"{tensors[idx].step :,}")
                matrix = tf.make_ndarray(tensors[idx].tensor_proto)
                x1, x2, y1, y2 = self.render_matrix(matrix,
                                                    axes[i, j],
                                                    color)
                x_min = min(x_min, x1)
                x_max = max(x_max, x2)
                y_min = min(y_min, y1)
                y_max = max(y_max, y2)

        for i in range(axes.shape[0]):
            for j in range(axes.shape[1]):
                axes[i, j].set_xlim(x_min, x_max)
                axes[i, j].set_ylim(y_min, y_max)



"""
Console colors
"""
from enum import Enum

_ANSI_CODES = dict(
    normal=0,
    bold=1,
    light=2,  # - PyCharm/Jupyter

    italic=3,  # - PyCharm/Jupyter
    underline=4,

    highlight=7,  # Changes background in PyCharm/Terminal

    # Colors
    black=30,
    red=31,
    green=32,
    orange=33,
    blue=34,
    purple=35,
    cyan=36,
    white=37,

    # Background [Not used anymore]
    bg_black=40,
    bg_red=41,
    bg_green=42,
    bg_orange=43,
    bg_blue=44,
    bg_purple=45,
    bg_cyan=46,
    bg_white=47,

    # Bright Colors [Not used anymore]
    bright_black=90,
    bright_red=91,
    bright_green=92,
    bright_orange=93,
    bright_blue=94,
    bright_purple=95,
    bright_cyan=96,
    bright_white=97,

    # Bright Background Colors [Not used anymore]
    bg_bright_black=100,
    bg_bright_red=101,
    bg_bright_green=102,
    bg_bright_orange=103,
    bg_bright_blue=104,
    bg_bright_purple=105,
    bg_bright_cyan=106,
    bg_bright_white=107,

    link=4,
)

ANSI_RESET = "\33[0m"

_HTML_STYLES = dict(
    normal=('', ''),
    bold=('<strong>', '</strong>'),
    underline=('<span style="text-decoration: underline">', '</span>'),
    light=('', ''),

    # Colors
    black=('<span style="color: #3E424D">', '</span>'),
    red=('<span style="color: #E75C58">', '</span>'),
    green=('<span style="color: #00A250">', '</span>'),
    orange=('<span style="color: #DDB62B">', '</span>'),
    blue=('<span style="color: #208FFB">', '</span>'),
    purple=('<span style="color: #D160C4">', '</span>'),
    cyan=('<span style="color: #60C6C8">', '</span>'),
    white=('<span style="color: #C5C1B4">', '</span>'),

    link=("<a href='{}' target='blank'>", "</a>"),
)


class StyleCode(Enum):
    r"""
    This is the base class for different style enumerations
    """

    def ansi(self):
        if self.value is None:
            return f"\33[{_ANSI_CODES['normal']}m"
        elif type(self.value) == str:
            return f"\33[{_ANSI_CODES[self.value]}m"
        elif type(self.value) == list:
            return ''.join([f"\33[{_ANSI_CODES[v]}m" for v in self.value])
        else:
            assert False

    def html_open(self, text: str):
        if self.value is None:
            return ""
        elif type(self.value) == str:
            return _HTML_STYLES[self.value][0].format(text)
        elif type(self.value) == list:
            return ''.join([_HTML_STYLES[v][0].format(text) for v in self.value])
        else:
            assert False

    def html_close(self, text):
        if self.value is None:
            return ""
        elif type(self.value) == str:
            return _HTML_STYLES[self.value][1].format(text)
        elif type(self.value) == list:
            return ''.join([_HTML_STYLES[v][1].format(text) for v in reversed(self.value)])
        else:
            assert False


def _test():
    for i in [0, 38, 48]:
        for j in [5]:
            for k in range(16):
                print("\33[{};{};{}m{:02d},{},{:03d}\33[0m\t".format(i, j, k, i, j, k),
                      end='')
                if (k + 1) % 6 == 0:
                    print("")
            print("")

    for i in range(0, 128):
        print(f"\33[{i}m{i :03d}\33[0m ", end='')
        if (i + 1) % 10 == 0:
            print("")

    print()

    print("▁▂▃▄▅▆▇█")
    print("▁▂▃▄▅▆▇█")


if __name__ == "__main__":
    _test()



import time
from pathlib import Path
from typing import List, Dict, Optional

import numpy as np

from labml import logger
from labml.internal.configs.processor import load_configs
from .. import util
from ..manage.runs import get_run_by_uuid, get_checkpoints
from ...logger import Text
from ...utils.notice import labml_notice


def struct_time_to_time(t: time.struct_time):
    return f"{t.tm_hour :02}:{t.tm_min :02}:{t.tm_sec :02}"


def struct_time_to_date(t: time.struct_time):
    return f"{t.tm_year :04}-{t.tm_mon :02}-{t.tm_mday :02}"


_GLOBAL_STEP = 'global_step'


class RunInfo:
    def __init__(self, *,
                 uuid: str,
                 python_file: str,
                 trial_date: str,
                 trial_time: str,
                 name: str,
                 comment: str,
                 commit: Optional[str] = None,
                 commit_message: Optional[str] = None,
                 is_dirty: bool = True,
                 experiment_path: Path,
                 start_step: int = 0,
                 notes: str = '',
                 load_run: Optional[str] = None,
                 tags: List[str]):
        self.name = name
        self.uuid = uuid
        self.commit = commit
        self.is_dirty = is_dirty
        self.python_file = python_file
        self.trial_date = trial_date
        self.trial_time = trial_time
        self.comment = comment
        self.commit_message = commit_message
        self.start_step = start_step

        self.load_run = load_run

        self.experiment_path = experiment_path
        self.run_path = experiment_path / str(uuid)
        self.pids_path = self.run_path / 'pids'
        self.checkpoint_path = self.run_path / "checkpoints"
        self.numpy_path = self.run_path / "numpy"

        self.diff_path = self.run_path / "source.diff"

        self.sqlite_path = self.run_path / "sqlite.db"
        self.artifacts_folder = self.run_path / "artifacts"
        self.tensorboard_log_path = self.run_path / "tensorboard"
        self.log_file = self.run_path / 'log.jsonl'

        self.info_path = self.run_path / "run.yaml"
        self.indicators_path = self.run_path / "indicators.yaml"
        self.configs_path = self.run_path / "configs.yaml"
        self.run_log_path = self.run_path / "run_log.jsonl"
        self.notes = notes
        self.tags = tags

    @classmethod
    def from_dict(cls, experiment_path: Path, data: Dict[str, any]):
        """
        ## Create a new trial from a dictionary
        """
        params = dict(experiment_path=experiment_path)
        params.update(data)
        return cls(**params)

    def to_dict(self):
        """
        ## Convert trial to a dictionary for saving
        """
        return dict(
            name=self.name,
            uuid=self.uuid,
            python_file=self.python_file,
            trial_date=self.trial_date,
            trial_time=self.trial_time,
            comment=self.comment,
            commit=self.commit,
            commit_message=self.commit_message,
            is_dirty=self.is_dirty,
            start_step=self.start_step,
            notes=self.notes,
            tags=self.tags,
            load_run=self.load_run
        )

    def pretty_print(self) -> List[str]:
        """
        ## 🎨 Pretty print trial for the python file header
        """

        # Trial information
        commit_status = "[dirty]" if self.is_dirty else "[clean]"
        res = [
            f"{self.trial_date} {self.trial_time}",
            self.comment,
            f"[{commit_status}]: {self.commit_message}",
            f"start_step: {self.start_step}"
        ]
        return res

    def __str__(self):
        return f"{self.__class__.__name__}(comment=\"{self.comment}\"," \
               f" commit=\"{self.commit_message}\"," \
               f" date={self.trial_date}, time={self.trial_time})"

    def __repr__(self):
        return self.__str__()

    def is_after(self, run: 'Run'):
        if run.trial_date < self.trial_date:
            return True
        elif run.trial_date > self.trial_date:
            return False
        elif run.trial_time < self.trial_time:
            return True
        else:
            return False


class Run(RunInfo):
    """
    # Trial 🏃‍

    Every trial in an experiment has same configs.
    It's just multiple runs.

    A new trial will replace checkpoints and TensorBoard summaries
    or previous trials, you should make a copy if needed.
    The performance log in `trials.yaml` is not replaced.

    You should run new trials after bug fixes or to see performance is
     consistent.

    If you want to try different configs, create multiple experiments.
    """

    diff: Optional[str]

    def __init__(self, *,
                 uuid: str,
                 python_file: str,
                 trial_date: str,
                 trial_time: str,
                 name: str,
                 comment: str,
                 commit: Optional[str] = None,
                 commit_message: Optional[str] = None,
                 is_dirty: bool = True,
                 experiment_path: Path,
                 start_step: int = 0,
                 notes: str = '',
                 tags: List[str]):
        super().__init__(python_file=python_file, trial_date=trial_date, trial_time=trial_time,
                         name=name, comment=comment, uuid=uuid, experiment_path=experiment_path,
                         commit=commit, commit_message=commit_message, is_dirty=is_dirty,
                         start_step=start_step, notes=notes, tags=tags)

    @classmethod
    def create(cls, *,
               uuid: str,
               experiment_path: Path,
               python_file: str,
               trial_time: time.struct_time,
               name: str,
               comment: str,
               tags: List[str]):
        """
        ## Create a new trial
        """
        return cls(python_file=python_file,
                   trial_date=struct_time_to_date(trial_time),
                   trial_time=struct_time_to_time(trial_time),
                   uuid=uuid,
                   experiment_path=experiment_path,
                   name=name,
                   comment=comment,
                   tags=tags)

    def make_path(self):
        run_path = Path(self.run_path)
        if not run_path.exists():
            try:
                run_path.mkdir(parents=True)
            except FileExistsError:
                pass


    def save_info(self):
        self.make_path()

        with open(str(self.info_path), "w") as file:
            file.write(util.yaml_dump(self.to_dict()))

        if self.diff is not None:
            with open(str(self.diff_path), "w") as f:
                f.write(self.diff)


def _get_run_checkpoint(run_path: Path, checkpoint: int = -1):
    checkpoints = list(get_checkpoints(run_path))
    if not checkpoints:
        return None

    checkpoint_steps = [int(c.name) for c in checkpoints]
    if checkpoint < 0:
        required_ci = np.max(checkpoint_steps) + checkpoint + 1
    else:
        required_ci = checkpoint

    checkpoint_steps = set(checkpoint_steps)

    for ci in range(required_ci, -1, -1):
        if ci not in checkpoint_steps:
            continue

        return ci


def get_run_checkpoint(run_uuid: str,
                       checkpoint: int = -1):
    run_path = get_run_by_uuid(run_uuid)
    if run_path is None:
        logger.log("Couldn't find a previous run")
        return None, None

    checkpoint = _get_run_checkpoint(run_path, checkpoint)

    if checkpoint is None:
        logger.log("Couldn't find checkpoints")
        return None, None

    logger.log(["Selected ",
                ("experiment", Text.key),
                " = ",
                (run_path.parent.name, Text.value),
                " ",
                ("run", Text.key),
                " = ",
                (run_uuid, Text.value),
                " ",
                ("checkpoint", Text.key),
                " = ",
                (str(checkpoint), Text.value)])

    checkpoint_path = run_path / "checkpoints"
    return checkpoint_path / str(checkpoint), checkpoint


def get_configs(run_uuid: str):
    run_path = get_run_by_uuid(run_uuid)
    if run_path is None:
        labml_notice(["Couldn't find a previous run to load configurations: ",
                      (run_uuid, Text.value)], is_danger=True)
        return None

    configs_path = run_path / "configs.yaml"
    configs = load_configs(configs_path)

    return configs



import signal
from typing import Optional, Tuple, Any, Collection

import typing

from labml import tracker, logger, experiment, monit
from labml.configs import BaseConfigs, meta_config, option
from labml.internal.monitor import Loop
from labml.logger import Text


class TrainingLoopIterator(Collection):
    def __init__(self, start: int, total: int, step: Optional[int]):
        self.step = step
        self.total = total
        self.start = start
        self.i = None

    def __iter__(self):
        self.i = None
        return self

    def __next__(self):
        if self.step is not None:
            if self.i is None:
                self.i = self.start
            else:
                self.i += self.step
        else:
            if self.i is None:
                self.i = 0
            else:
                self.i += 1

        if self.i >= self.total:
            raise StopIteration()

        if self.step is None:
            return tracker.get_global_step()
        else:
            return self.i

    def __len__(self) -> int:
        if self.step is not None:
            return (self.total - self.start) // self.step
        else:
            return self.total

    def __contains__(self, x: object) -> bool:
        return False


class TrainingLoop:
    __loop: Loop
    __signal_received: Optional[Tuple[Any, Any]]

    def __init__(self, *,
                 loop_count: int,
                 loop_step: Optional[int],
                 is_save_models: bool,
                 log_new_line_interval: int,
                 log_write_interval: int,
                 save_models_interval: int,
                 is_loop_on_interrupt: bool):
        self.__loop_count = loop_count
        self.__loop_step = loop_step
        self.__is_save_models = is_save_models
        self.__log_new_line_interval = log_new_line_interval
        self.__log_write_interval = log_write_interval
        self.__last_write_step = 0
        self.__last_new_line_step = 0
        self.__save_models_interval = save_models_interval
        self.__last_save_step = 0
        self.__signal_received = None
        self.__is_loop_on_interrupt = is_loop_on_interrupt
        self._iter = None

    def __iter__(self):
        self._iter = TrainingLoopIterator(tracker.get_global_step(),
                                          self.__loop_count,
                                          self.__loop_step)

        self.__loop = monit.loop(typing.cast(Collection, self._iter))

        iter(self.__loop)
        try:
            self.old_handler = signal.signal(signal.SIGINT, self.__handler)
        except ValueError:
            pass
        return self

    def __finish(self):
        try:
            signal.signal(signal.SIGINT, self.old_handler)
        except ValueError:
            pass
        tracker.save()
        tracker.new_line()
        if self.__is_save_models:
            logger.log("Saving model...")
            experiment.save_checkpoint()

    # def is_interval(self, interval: int, global_step: Optional[int] = None):
    #     if global_step is None:
    #         global_step = tracker.get_global_step()
    #
    #     if global_step - self.__loop_step < 0:
    #         return False
    #
    #     if global_step // interval > (global_step - self.__loop_step) // interval:
    #         return True
    #     else:
    #         return False

    def __next__(self):
        if self.__signal_received is not None:
            logger.log('\nKilling Loop.', Text.danger)
            monit.finish_loop()
            self.__finish()
            raise StopIteration("SIGINT")

        try:
            global_step = next(self.__loop)
        except StopIteration as e:
            self.__finish()
            raise e

        tracker.set_global_step(global_step)

        if global_step - self.__last_write_step >= self.__log_write_interval:
            tracker.save()
            self.__last_write_step = global_step
        if global_step - self.__last_new_line_step >= self.__log_new_line_interval:
            tracker.new_line()
            self.__last_new_line_step = global_step
        # if self.is_interval(self.__log_write_interval, global_step):
        #     tracker.save()
        # if self.is_interval(self.__log_new_line_interval, global_step):
        #     logger.log()

        # if (self.__is_save_models and
        #         self.is_interval(self.__save_models_interval, global_step)):
        #     experiment.save_checkpoint()
        if (self.__is_save_models and
                global_step - self.__last_save_step >= self.__save_models_interval):
            experiment.save_checkpoint()

        return global_step

    def __handler(self, sig, frame):
        # Pass second interrupt without delaying
        if self.__signal_received is not None:
            logger.log('\nSIGINT received twice. Stopping...', Text.danger)
            self.old_handler(*self.__signal_received)
            return

        if self.__is_loop_on_interrupt:
            # Store the interrupt signal for later
            self.__signal_received = (sig, frame)
            logger.log('\nSIGINT received. Delaying KeyboardInterrupt.', Text.danger)
        else:
            self.__finish()
            logger.log('Killing loop...', Text.danger)
            self.old_handler(sig, frame)

    def __str__(self):
        return "LabTrainingLoop"


class TrainingLoopConfigs(BaseConfigs):
    loop_count: int = 10
    loop_step: int = 1
    is_save_models: bool = False
    log_new_line_interval: int = 1
    log_write_interval: int = 1
    save_models_interval: int = 1
    is_loop_on_interrupt: bool = False

    training_loop: TrainingLoop


@option(TrainingLoopConfigs.training_loop)
def _loop_configs(c: TrainingLoopConfigs):
    return TrainingLoop(loop_count=c.loop_count,
                        loop_step=c.loop_step,
                        is_save_models=c.is_save_models,
                        log_new_line_interval=c.log_new_line_interval,
                        log_write_interval=c.log_write_interval,
                        save_models_interval=c.save_models_interval,
                        is_loop_on_interrupt=c.is_loop_on_interrupt)


meta_config(TrainingLoopConfigs.loop_step,
            TrainingLoopConfigs.loop_count,
            TrainingLoopConfigs.is_save_models,
            TrainingLoopConfigs.log_new_line_interval,
            TrainingLoopConfigs.log_write_interval,
            TrainingLoopConfigs.save_models_interval,
            TrainingLoopConfigs.is_loop_on_interrupt)



import time
import webbrowser
from typing import Dict, Optional

from labml import logger
from labml.internal.api import ApiCaller, Packet
from labml.logger import Text

from ..configs.processor import ConfigsSaver


class WebApiConfigsSaver(ConfigsSaver):
    def __init__(self, api_experiment: 'ApiExperiment'):
        self.api_experiment = api_experiment

    def save(self, configs: Dict):
        self.api_experiment.save_configs(configs)


class ApiExperiment:
    configs_saver: Optional[WebApiConfigsSaver]

    def __init__(self, api_caller: ApiCaller, *,
                 frequency: float,
                 open_browser: bool):
        super().__init__()

        self.frequency = frequency
        self.open_browser = open_browser
        self.api_caller = api_caller
        self.run_uuid = None
        self.name = None
        self.comment = None
        self.state = None
        self.configs_saver = None
        self.api_caller.add_state_attribute('configs')

    def set_info(self, *,
                 run_uuid: str,
                 name: str,
                 comment: str):
        self.run_uuid = run_uuid
        self.name = name
        self.comment = comment

    def get_configs_saver(self):
        if self.configs_saver is None:
            self.configs_saver = WebApiConfigsSaver(self)
        return self.configs_saver

    def save_configs(self, configs: Dict[str, any]):
        self.api_caller.push(Packet({'configs': configs}))

    def start(self):
        data = {
            'name': self.name,
            'comment': self.comment,
            'time': time.time()
        }

        self.api_caller.push(Packet(data, callback=self._started))

        from labml.internal.api.logs import API_LOGS
        API_LOGS.set_api(self.api_caller, frequency=self.frequency)

    def _started(self, url):
        if url is None:
            return None

        logger.log([('Monitor experiment at ', Text.meta), (url, Text.link)])
        if self.open_browser:
            webbrowser.open(url)

    def status(self, rank: int, status: str, details: str, time_: float):
        self.state = {
            'rank': rank,
            'status': status,
            'details': details,
            'time': time_
        }

        self.api_caller.push(Packet({
            'status': self.state,
            'time': time.time()
        }))

        # TODO: Will have to fix this when there are other statuses that dont stop the experiment
        # This will stop the thread after sending all the data
        self.api_caller.stop()



from typing import Optional, List

from .destinations.factory import create_destination
from .inspect import Inspect
from .types import LogPart
from ..api.logs import API_LOGS as _


class Logger:
    """
    This handles the interactions among sections, loop and store
    """

    def __init__(self):
        self.__destination = create_destination()
        self.__inspect = Inspect(self)

    def log(self, parts: List[LogPart], *,
            is_new_line: bool = True,
            is_reset: bool = True):
        self.__destination.log(parts, is_new_line=is_new_line, is_reset=is_reset)

    def info(self, *args, **kwargs):
        self.__inspect.info(*args, **kwargs)


_internal: Optional[Logger] = None


def logger_singleton() -> Logger:
    global _internal
    if _internal is None:
        _internal = Logger()

    return _internal



"""
---
title: Configurable Transformer Components
summary: These are configurable components that can be re-used quite easily.
---

# Configurable Transformer Components
"""
import copy

import torch.nn as nn

from labml.configs import BaseConfigs, option, calculate
from labml_helpers.module import Module
from .mha import MultiHeadAttention
from .models import EmbeddingsWithPositionalEncoding, EmbeddingsWithLearnedPositionalEncoding, FeedForward, \
    TransformerLayer, Encoder, Decoder, Generator, EncoderDecoder


class TransformerConfigs(BaseConfigs):
    n_heads: int = 8
    d_model: int = 512
    n_layers: int = 6
    d_ff: int = 2048
    dropout: float = 0.1
    n_src_vocab: int
    n_tgt_vocab: int

    encoder_attn: MultiHeadAttention = 'mha'
    decoder_attn: MultiHeadAttention = 'mha'
    decoder_mem_attn: MultiHeadAttention = 'mha'
    feed_forward: FeedForward

    encoder_layer: TransformerLayer = 'normal'
    decoder_layer: TransformerLayer = 'normal'

    encoder: Encoder = 'normal'
    decoder: Decoder = 'normal'

    src_embed: Module = 'fixed_pos'
    tgt_embed: Module = 'fixed_pos'

    generator: Generator = 'default'

    encoder_decoder: EncoderDecoder


@option(TransformerConfigs.feed_forward, 'default')
def _feed_forward(c: TransformerConfigs):
    return FeedForward(c.d_model, c.d_ff, c.dropout)


### MHA
def _mha(c: TransformerConfigs):
    return MultiHeadAttention(c.n_heads, c.d_model)


calculate(TransformerConfigs.encoder_attn, 'mha', _mha)
calculate(TransformerConfigs.decoder_attn, 'mha', _mha)
calculate(TransformerConfigs.decoder_mem_attn, 'mha', _mha)


### Relative MHA
def _relative_mha(c: TransformerConfigs):
    from .relative_mha import RelativeMultiHeadAttention
    return RelativeMultiHeadAttention(c.n_heads, c.d_model)


calculate(TransformerConfigs.encoder_attn, 'relative', _relative_mha)
calculate(TransformerConfigs.decoder_attn, 'relative', _relative_mha)
calculate(TransformerConfigs.decoder_mem_attn, 'relative', _relative_mha)


@option(TransformerConfigs.encoder_layer, 'normal')
def _encoder_layer(c: TransformerConfigs):
    return TransformerLayer(d_model=c.d_model, self_attn=c.encoder_attn,
                            src_attn=None, feed_forward=copy.deepcopy(c.feed_forward),
                            dropout_prob=c.dropout)


@option(TransformerConfigs.decoder_layer, 'normal')
def _decoder_layer(c: TransformerConfigs):
    return TransformerLayer(d_model=c.d_model, self_attn=c.decoder_attn,
                            src_attn=c.decoder_mem_attn, feed_forward=copy.deepcopy(c.feed_forward),
                            dropout_prob=c.dropout)


@option(TransformerConfigs.encoder, 'normal')
def _encoder(c: TransformerConfigs):
    return Encoder(c.encoder_layer, c.n_layers)


@option(TransformerConfigs.decoder, 'normal')
def _decoder(c: TransformerConfigs):
    return Decoder(c.decoder_layer, c.n_layers)


@option(TransformerConfigs.generator, 'default')
def _generator(c: TransformerConfigs):
    return Generator(c.n_tgt_vocab, c.d_model)


### Positional Embeddings
@option(TransformerConfigs.src_embed, 'fixed_pos')
def _src_embed_with_positional(c: TransformerConfigs):
    return EmbeddingsWithPositionalEncoding(c.d_model, c.n_src_vocab)


@option(TransformerConfigs.tgt_embed, 'fixed_pos')
def _tgt_embed_with_positional(c: TransformerConfigs):
    return EmbeddingsWithPositionalEncoding(c.d_model, c.n_tgt_vocab)


### Learned Positional Embeddings
@option(TransformerConfigs.src_embed, 'learned_pos')
def _src_embed_with_learned_positional(c: TransformerConfigs):
    return EmbeddingsWithLearnedPositionalEncoding(c.d_model, c.n_src_vocab)


@option(TransformerConfigs.tgt_embed, 'learned_pos')
def _tgt_embed_with_learned_positional(c: TransformerConfigs):
    return EmbeddingsWithLearnedPositionalEncoding(c.d_model, c.n_tgt_vocab)


### No Positional Embeddings
@option(TransformerConfigs.src_embed, 'no_pos')
def _src_embed_without_positional(c: TransformerConfigs):
    return nn.Embedding(c.n_src_vocab, c.d_model)


@option(TransformerConfigs.tgt_embed, 'no_pos')
def _tgt_embed_without_positional(c: TransformerConfigs):
    return nn.Embedding(c.n_tgt_vocab, c.d_model)


@option(TransformerConfigs.encoder_decoder, 'normal')
def _encoder_decoder(c: TransformerConfigs):
    return EncoderDecoder(c.encoder, c.decoder, c.src_embed, c.tgt_embed, c.generator)



"""
---
title: Label Smoothing Loss
summary: >
  This is an implementation of label smoothing loss, that can be used as
  an alternative to cross entropy loss for improved accuracy.
---

# Label Smoothing Loss
"""
import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn as nn

from labml_helpers.module import Module


class LabelSmoothingLoss(Module):
    def __init__(self, size: int, padding_idx: int, smoothing: float = 0.0):
        super().__init__()
        self.loss = nn.KLDivLoss(reduction='sum')
        self.padding_idx = padding_idx
        self.confidence = 1.0 - smoothing
        self.smoothing = smoothing
        self.size = size
        self.true_dist = None

    def __call__(self, x: torch.Tensor, target: torch.Tensor):
        assert x.shape[1] == self.size
        true_dist = x.clone()
        true_dist.fill_(self.smoothing / (self.size - 2))
        true_dist.scatter_(1, target.unsqueeze(1), self.confidence)
        true_dist[:, self.padding_idx] = 0
        mask = torch.nonzero(target == self.padding_idx, as_tuple=False)
        if mask.dim() > 0:
            true_dist.index_fill_(0, mask.squeeze(), 0.0)
        self.true_dist = true_dist
        return self.loss(x, true_dist.detach())


def _test_label_smoothing():
    smooth_loss = LabelSmoothingLoss(5, 0, 0.4)
    predict = torch.tensor([[0, 0.2, 0.7, 0.1, 0],
                            [0, 0.2, 0.7, 0.1, 0],
                            [0, 0.2, 0.7, 0.1, 0]], dtype=torch.float)
    _ = smooth_loss(predict.log(),
                    torch.tensor([2, 1, 0], dtype=torch.long))

    # Show the target distributions expected by the system.
    plt.imshow(smooth_loss.true_dist)
    plt.show()

    smooth_loss = LabelSmoothingLoss(5, 0, 0.1)

    def loss_sample(x):
        d = x + 3 * 1
        predict2 = torch.tensor([[0, x / d, 1 / d, 1 / d, 1 / d],
                                 ], dtype=torch.float)
        # print(predict)
        return smooth_loss(predict2.log(),
                           torch.tensor([1], dtype=torch.long)).item()

    plt.plot(np.arange(1, 100), [loss_sample(x) for x in range(1, 100)])
    plt.show()


if __name__ == '__main__':
    _test_label_smoothing()



import numpy as np

try:
    import torch
except ImportError:
    torch = None


def to_numpy(value):
    if isinstance(value, int) or isinstance(value, float):
        return np.array(value)
    elif isinstance(value, np.number):
        return np.array(value.item())
    elif isinstance(value, list):
        return np.array(value)
    elif isinstance(value, np.ndarray):
        return value
    elif torch is not None:
        if isinstance(value, torch.nn.parameter.Parameter):
            return value.data.cpu().numpy()
        elif isinstance(value, torch.Tensor):
            return value.data.cpu().numpy()

    raise ValueError(f"Unknown type {type(value)}")



from typing import List, Tuple, Union

import torch


def detach(s: Union[Tuple, List, torch.Tensor]):
    if s is None:
        return None
    elif isinstance(s, torch.Tensor):
        return s.detach()
    elif isinstance(s, tuple):
        return tuple(detach(e) for e in s)
    else:
        return list(detach(e) for e in s)



import datetime

import torch.distributed
from torch.nn.parallel import DistributedDataParallel

from labml import experiment, monit
from labml.configs import option
from labml.logger import inspect
from labml_samples.pytorch.mnist.e_labml_helpers import Configs, Net


@option(Configs.model)
def ddp_model(c: Configs):
    return DistributedDataParallel(Net().to(c.device), device_ids=[c.device])


def main(local_rank, rank, world_size, uuid, init_method: str = 'tcp://localhost:23456'):
    with monit.section('Distributed'):
        torch.distributed.init_process_group("gloo",
                                             timeout=datetime.timedelta(seconds=30),
                                             init_method=init_method,
                                             rank=rank,
                                             world_size=world_size)
    conf = Configs()
    experiment.create(uuid=uuid, name='mnist ddp')
    experiment.distributed(local_rank, world_size)
    experiment.configs(conf,
                       {'optimizer.optimizer': 'Adam',
                        'optimizer.learning_rate': 1e-4,
                        'model': 'ddp_model',
                        'device.cuda_device': local_rank})
    conf.set_seed.set()
    experiment.add_pytorch_models(dict(model=conf.model))
    with experiment.start():
        conf.run()


def _launcher():
    import os
    world_size = int(os.environ['WORLD_SIZE'])
    run_uuid = os.environ['RUN_UUID']
    local_rank = int(os.environ['LOCAL_RANK'])
    rank = int(os.environ['RANK'])
    inspect(world_size=os.environ['WORLD_SIZE'],
            run_uuid=os.environ['RUN_UUID'],
            local_rank=os.environ['LOCAL_RANK'],
            rank=os.environ['RANK'],
            master_addr=os.environ['MASTER_ADDR'],
            master_port=os.environ['MASTER_PORT'])
    main(local_rank, rank, world_size, run_uuid, 'env://')


if __name__ == '__main__':
    # Run single GPU
    # main(0, 1, experiment.generate_uuid())

    # Spawn multiple GPU
    # torch.multiprocessing.spawn(main, args=(3, experiment.generate_uuid()), nprocs=3, join=True)

    _launcher()



import functools
from typing import Iterable, Sized, Collection, Callable
from typing import Union, Optional, overload

from labml.internal.monitor import monitor_singleton as _internal


def clear():
    _internal().clear()


def func(name, *,
         is_silent: bool = False,
         is_timed: bool = True,
         is_partial: bool = False,
         is_new_line: bool = True,
         is_children_silent: bool = False,
         total_steps: float = 1.0):
    def decorator_func(f: Callable):
        def wrapper(*args, **kwargs):
            with section(name,
                         is_silent=is_silent,
                         is_timed=is_timed,
                         is_partial=is_partial,
                         is_new_line=is_new_line,
                         is_children_silent=is_children_silent,
                         total_steps=total_steps):
                return f(*args, **kwargs)

        return wrapper

    return decorator_func


def iterate(name, iterable: Union[Iterable, Sized, int],
            total_steps: Optional[int] = None, *,
            is_silent: bool = False,
            is_children_silent: bool = False,
            is_timed: bool = True,
            context=None):
    return _internal().iterate(name, iterable, total_steps,
                               is_silent=is_silent,
                               is_children_silent=is_children_silent,
                               is_timed=is_timed,
                               section=context)


def enum(name, iterable: Sized, *,
         is_silent: bool = False,
         is_children_silent: bool = False,
         is_timed: bool = True,
         context=None):
    return _internal().enum(name, iterable,
                            is_silent=is_silent,
                            is_children_silent=is_children_silent,
                            is_timed=is_timed,
                            section=context)


def section(name, *,
            is_silent: bool = False,
            is_timed: bool = True,
            is_partial: bool = False,
            is_new_line: bool = True,
            is_children_silent: bool = False,
            total_steps: float = 1.0):
    return _internal().section(name, is_silent=is_silent,
                               is_timed=is_timed,
                               is_partial=is_partial,
                               total_steps=total_steps,
                               is_new_line=is_new_line,
                               is_children_silent=is_children_silent)


def progress(steps: float):
    _internal().progress(steps)


def fail():
    _internal().set_successful(False)


@overload
def loop(iterator_: int, *,
         is_track: bool = True,
         is_print_iteration_time: bool = True):
    ...


@overload
def loop(iterator_: range, *,
         is_track: bool = True,
         is_print_iteration_time: bool = True):
    ...


@overload
def loop(iterator_: Collection, *,
         is_track: bool = True,
         is_print_iteration_time: bool = True):
    ...


def loop(iterator_: Union[Collection, range, int], *,
         is_track: bool = True,
         is_print_iteration_time: bool = True):
    """
        This has multiple overloads

        .. function:: loop(iterator_: range, *, is_track=True, is_print_iteration_time=True)
            :noindex:

        .. function:: loop(iterator_: int, *, is_track=True, is_print_iteration_time=True)
            :noindex:
        """

    if type(iterator_) == int:
        return _internal().loop(range(iterator_),
                                is_track=is_track,
                                is_print_iteration_time=is_print_iteration_time)
    else:
        return _internal().loop(iterator_,
                                is_track=is_track,
                                is_print_iteration_time=is_print_iteration_time)


def finish_loop():
    _internal().finish_loop()



from typing import Any

import torch.nn


class Module(torch.nn.Module):
    r"""
    Wraps ``torch.nn.Module`` to overload ``__call__`` instead of
    ``forward`` for better type checking.
    
    `PyTorch Github issue for clarification <https://github.com/pytorch/pytorch/issues/44605>`_
    """

    def _forward_unimplemented(self, *input: Any) -> None:
        # To stop PyTorch from giving abstract methods warning
        pass

    def __init_subclass__(cls, **kwargs):
        if cls.__dict__.get('__call__', None) is None:
            return

        setattr(cls, 'forward', cls.__dict__['__call__'])
        delattr(cls, '__call__')

    @property
    def device(self):
        params = self.parameters()
        try:
            sample_param = next(params)
            return sample_param.device
        except StopIteration:
            raise RuntimeError(f"Unable to determine"
                               f" device of {self.__class__.__name__}") from None


if __name__ == '__main__':
    m = Module()
    print(m.device)



"""
---
title: Transformers
summary: >
  This is a collection of PyTorch implementations/tutorials of
  transformers and related techniques.
---

# Transformers

## Transformer Building Blocks

* [Multi-head attention](mha.html)
* [Relative multi-head attention](relative_mha.html)
* [Transformer Encoder and Decoder Models](models.html)
* [Fixed positional encoding](positional_encoding.html)

## [kNN-LM](knn)

This is an implementation of the paper
 [Generalization through Memorization: Nearest Neighbor Language Models](https://arxiv.org/abs/1911.00172).
"""

from .configs import TransformerConfigs
from .models import TransformerLayer, Encoder, Decoder, Generator, EncoderDecoder
from .mha import MultiHeadAttention
from .relative_mha import RelativeMultiHeadAttention



from typing import List

import altair as alt
from labml.internal.analytics.altair.utils import TABLEAU_10

try:
    import torch
except ImportError:
    torch = None


def data_to_table(data, step):
    table = []

    if torch is not None and isinstance(data, torch.Tensor):
        data = data.detach().cpu().numpy()

    for i in range(data.shape[0]):
        if len(data.shape) == 2:
            m = data.shape[1] // 2
            row = {'step': data[i, 0], 'v5': data[i, m]}
            for j in range(1, min(m, 5)):
                row[f"v{5 - j}"] = data[i, m - j]
                row[f"v{5 + j}"] = data[i, m + j]
        elif step is not None:
            row = {'step': step[i],
                   'v5': data[i]}
        else:
            row = {'step': i,
                   'v5': data[i]}
        table.append(row)

    return alt.Data(values=table)


def _render_density(table: alt.Data, *,
                    name: str,
                    x_name: str,
                    line_color: str,
                    range_color: str,
                    levels: int,
                    alpha: float,
                    selection: alt.Selection = None,
                    x_scale: alt.Scale = alt.Undefined,
                    y_scale: alt.Scale = alt.Undefined) -> alt.Chart:
    areas: List[alt.Chart] = []
    for i in range(1, levels):
        y = f"v{5 - i}:Q"
        y2 = f"v{5 + i}:Q"

        areas.append(
            alt.Chart(table)
                .mark_area(opacity=alpha ** i)
                .encode(x=alt.X('step:Q', scale=x_scale),
                        y=alt.Y(y, scale=y_scale),
                        y2=alt.Y2(y2),
                        color=alt.value(range_color)
                        )
        )

    line: alt.Chart = (
        alt.Chart(table)
            .mark_line()
            .encode(x=alt.X('step:Q', scale=x_scale, title=x_name),
                    y=alt.Y("v5:Q", scale=y_scale, title=name),
                    color=alt.value(line_color)
                    )
    )
    if selection is not None:
        line = line.add_selection(selection)

    areas_sum = None
    for a in areas:
        if areas_sum is None:
            areas_sum = a
        else:
            areas_sum += a

    if areas_sum is None:
        return line
    else:
        return areas_sum + line


def render(tables: List[alt.Data], *,
           names: List[str],
           is_independent: bool,
           levels=5,
           alpha=0.6,
           height: int,
           width: int,
           height_minimap: int):
    zoom = alt.selection_interval(encodings=["x", "y"])

    minimaps = None
    for i, t in enumerate(tables):
        z = zoom if i == 0 else None
        minimap = _render_density(t,
                                  name='',
                                  x_name='',
                                  line_color=TABLEAU_10[i % 10],
                                  range_color=TABLEAU_10[i % 10],
                                  levels=levels,
                                  alpha=alpha,
                                  selection=z)
        if minimaps is None:
            minimaps = minimap
        else:
            minimaps += minimap

    details = None
    for i, t in enumerate(tables):
        detail = _render_density(t,
                                 name=names[i] if len(names) == 1 else '',
                                 x_name='Step',
                                 line_color=TABLEAU_10[i % 10],
                                 range_color=TABLEAU_10[i % 10],
                                 levels=levels,
                                 alpha=alpha,
                                 x_scale=alt.Scale(domain={'selection': zoom.name,
                                                           "encoding": "x"}),
                                 y_scale=alt.Scale(domain={'selection': zoom.name,
                                                           "encoding": "y"}))
        if details is None:
            details = detail
        elif is_independent:
            details = alt.layer(details, detail).resolve_scale(y='independent')
        else:
            details += detail

    colors = alt.Data(values=[{'idx': i, 'name': name} for i, name in enumerate(names)])
    legend = alt.Chart(colors).mark_rect().encode(
        alt.Y('name:N',
              sort=alt.EncodingSortField(field='i', order='ascending'),
              axis=alt.Axis(
                  orient='right',
                  titleX=7,
                  titleY=-2,
                  titleAlign='left',
                  titleAngle=0
              ),
              title=''),
        alt.Color('idx:O',
                  scale=alt.Scale(domain=list(range(len(tables))),
                                  range=TABLEAU_10),
                  legend=None))

    minimaps = minimaps.properties(width=width, height=height_minimap)
    details = details.properties(width=width, height=height)

    if len(names) == 1:
        return details & minimaps
    else:
        return (details & minimaps) | legend



from typing import Dict

import numpy as np
from .numeric import Queue, Histogram, Scalar

try:
    import torch
except ImportError:
    torch = None


def load_indicator_from_dict(data: Dict[str, any]):
    class_name = data['class_name']
    del data['class_name']

    if class_name == 'Queue':
        return Queue(**data)
    elif class_name == 'Histogram':
        return Histogram(**data)
    elif class_name == 'Scalar':
        return Scalar(**data)
    else:
        raise ValueError(f"Unknown indicator: {class_name}")


def create_default_indicator(name: str, value: any, is_print: bool):
    if isinstance(value, int) or isinstance(value, float):
        return Scalar(name, is_print)
    elif isinstance(value, np.number):
        return Scalar(name, is_print)
    elif isinstance(value, list):
        return Scalar(name, is_print)
    elif isinstance(value, np.ndarray):
        return Scalar(name, is_print)
    elif torch is not None:
        from torch.optim.optimizer import Optimizer
        if isinstance(value, torch.nn.parameter.Parameter):
            return Scalar(name, is_print)
        elif isinstance(value, torch.Tensor):
            return Scalar(name, is_print)
        elif isinstance(value, torch.nn.Module):
            from labml.internal.tracker.indicators.aggregate import PyTorchModule
            return PyTorchModule(name, is_print)
        elif isinstance(value, Optimizer):
            from labml.internal.tracker.indicators.aggregate import PyTorchOptimizer
            return PyTorchOptimizer(name, is_print)
        elif isinstance(value, tuple) and isinstance(value[0], Optimizer):
            from labml.internal.tracker.indicators.aggregate import PyTorchOptimizer
            return PyTorchOptimizer(name, is_print)

    raise ValueError(f"Unknown type {type(value)}")



from pathlib import PurePath
from typing import Dict, Optional


def new_run(python_file: PurePath, *,
            configs: Optional[Dict[str, any]] = None,
            comment: Optional[str] = None):
    from labml.internal.manage.process import new_run as _new_run
    _new_run(python_file, configs=configs, comment=comment)


def new_run_process(python_file: PurePath, *,
                    configs: Optional[Dict[str, any]] = None,
                    comment: Optional[str] = None):
    new_run_process(python_file, configs=configs, comment=comment)



import argparse

from labml import logger
from labml.logger import Text


def _open_dashboard():
    try:
        import labml_dashboard
    except (ImportError, ModuleNotFoundError):
        logger.log("Cannot import ", ('labml_dashboard', Text.highlight), '.')
        logger.log('Install with ',
                   ('pip install labml_dashboard', Text.value))
        return

    labml_dashboard.start_server()


def main():
    parser = argparse.ArgumentParser(description='LabML CLI')
    parser.add_argument('command', choices=['dashboard'])

    args = parser.parse_args()

    if args.command == 'dashboard':
        _open_dashboard()
    else:
        raise ValueError('Unknown command', args.command)



from typing import List, Union, Tuple, Optional

from labml.internal.util.colors import StyleCode


class Destination:
    def log(self, parts: List[Union[str, Tuple[str, Optional[StyleCode]]]], *,
            is_new_line: bool,
            is_reset: bool):
        raise NotImplementedError()



"""
---
title: Proximal Policy Optimization (PPO)
summary: >
 An annotated implementation of Proximal Policy Optimization (PPO) algorithm in PyTorch.
---

# Proximal Policy Optimization (PPO)

This is a an implementation of [Proximal Policy Optimization - PPO](https://arxiv.org/abs/1707.06347).

You can find an experiment that uses it [here](experiment.html).
The experiment uses [Generalized Advantage Estimation](gae.html).
"""

import torch

from labml_helpers.module import Module
from labml_nn.rl.ppo.gae import GAE


class ClippedPPOLoss(Module):
    """
    ## PPO Loss

    We want to maximize policy reward
     $$\max_\theta J(\pi_\theta) =
       \mathop{\mathbb{E}}_{\tau \sim \pi_\theta}\Biggl[\sum_{t=0}^\infty \gamma^t r_t \Biggr]$$
     where $r$ is the reward, $\pi$ is the policy, $\tau$ is a trajectory sampled from policy,
     and $\gamma$ is the discount factor between $[0, 1]$.

    \begin{align}
    \mathbb{E}_{\tau \sim \pi_\theta} \Biggl[
     \sum_{t=0}^\infty \gamma^t A^{\pi_{OLD}}(s_t, a_t)
    \Biggr] &=
    \\
    \mathbb{E}_{\tau \sim \pi_\theta} \Biggl[
      \sum_{t=0}^\infty \gamma^t \Bigl(
       Q^{\pi_{OLD}}(s_t, a_t) - V^{\pi_{OLD}}(s_t)
      \Bigr)
     \Biggr] &=
    \\
    \mathbb{E}_{\tau \sim \pi_\theta} \Biggl[
      \sum_{t=0}^\infty \gamma^t \Bigl(
       r_t + V^{\pi_{OLD}}(s_{t+1}) - V^{\pi_{OLD}}(s_t)
      \Bigr)
     \Biggr] &=
    \\
    \mathbb{E}_{\tau \sim \pi_\theta} \Biggl[
      \sum_{t=0}^\infty \gamma^t \Bigl(
       r_t
      \Bigr)
     \Biggr]
     - \mathbb{E}_{\tau \sim \pi_\theta}
        \Biggl[V^{\pi_{OLD}}(s_0)\Biggr] &=
    J(\pi_\theta) - J(\pi_{\theta_{OLD}})
    \end{align}

    So,
     $$\max_\theta J(\pi_\theta) =
       \max_\theta \mathbb{E}_{\tau \sim \pi_\theta} \Biggl[
          \sum_{t=0}^\infty \gamma^t A^{\pi_{OLD}}(s_t, a_t)
       \Biggr]$$

    Define discounted-future state distribution,
     $$d^\pi(s) = (1 - \gamma) \sum_{t=0}^\infty \gamma^t P(s_t = s | \pi)$$

    Then,
    \begin{align}
    J(\pi_\theta) - J(\pi_{\theta_{OLD}})
    &= \mathbb{E}_{\tau \sim \pi_\theta} \Biggl[
     \sum_{t=0}^\infty \gamma^t A^{\pi_{OLD}}(s_t, a_t)
    \Biggr]
    \\
    &= \frac{1}{1 - \gamma}
     \mathbb{E}_{s \sim d^{\pi_\theta}, a \sim \pi_\theta} \Bigl[
      A^{\pi_{OLD}}(s, a)
     \Bigr]
    \end{align}

    Importance sampling $a$ from $\pi_{\theta_{OLD}}$,

    \begin{align}
    J(\pi_\theta) - J(\pi_{\theta_{OLD}})
    &= \frac{1}{1 - \gamma}
     \mathbb{E}_{s \sim d^{\pi_\theta}, a \sim \pi_\theta} \Bigl[
      A^{\pi_{OLD}}(s, a)
     \Bigr]
    \\
    &= \frac{1}{1 - \gamma}
     \mathbb{E}_{s \sim d^{\pi_\theta}, a \sim \pi_{\theta_{OLD}}} \Biggl[
      \frac{\pi_\theta(a|s)}{\pi_{\theta_{OLD}}(a|s)} A^{\pi_{OLD}}(s, a)
     \Biggr]
    \end{align}

    Then we assume $d^\pi_\theta(s)$ and  $d^\pi_{\theta_{OLD}}(s)$ are similar.
    The error we introduce to $J(\pi_\theta) - J(\pi_{\theta_{OLD}})$
     by this assumtion is bound by the KL divergence between
     $\pi_\theta$ and $\pi_{\theta_{OLD}}$.
    [Constrained Policy Optimization](https://arxiv.org/abs/1705.10528)
     shows the proof of this. I haven't read it.


    \begin{align}
    J(\pi_\theta) - J(\pi_{\theta_{OLD}})
    &= \frac{1}{1 - \gamma}
     \mathop{\mathbb{E}}_{s \sim d^{\pi_\theta} \atop a \sim \pi_{\theta_{OLD}}} \Biggl[
      \frac{\pi_\theta(a|s)}{\pi_{\theta_{OLD}}(a|s)} A^{\pi_{OLD}}(s, a)
     \Biggr]
    \\
    &\approx \frac{1}{1 - \gamma}
     \mathop{\mathbb{E}}_{\color{orange}{s \sim d^{\pi_{\theta_{OLD}}}}
     \atop a \sim \pi_{\theta_{OLD}}} \Biggl[
      \frac{\pi_\theta(a|s)}{\pi_{\theta_{OLD}}(a|s)} A^{\pi_{OLD}}(s, a)
     \Biggr]
    \\
    &= \frac{1}{1 - \gamma} \mathcal{L}^{CPI}
    \end{align}
    """

    def __init__(self):
        super().__init__()

    def __call__(self, log_pi: torch.Tensor, sampled_log_pi: torch.Tensor,
                 advantage: torch.Tensor, clip: float) -> torch.Tensor:
        # ratio $r_t(\theta) = \frac{\pi_\theta (a_t|s_t)}{\pi_{\theta_{OLD}} (a_t|s_t)}$;
        # *this is different from rewards* $r_t$.
        ratio = torch.exp(log_pi - sampled_log_pi)

        # \begin{align}
        # \mathcal{L}^{CLIP}(\theta) =
        #  \mathbb{E}_{a_t, s_t \sim \pi_{\theta{OLD}}} \biggl[
        #    min \Bigl(r_t(\theta) \bar{A_t},
        #              clip \bigl(
        #               r_t(\theta), 1 - \epsilon, 1 + \epsilon
        #              \bigr) \bar{A_t}
        #    \Bigr)
        #  \biggr]
        # \end{align}
        #
        # The ratio is clipped to be close to 1.
        # We take the minimum so that the gradient will only pull
        # $\pi_\theta$ towards $\pi_{\theta_{OLD}}$ if the ratio is
        # not between $1 - \epsilon$ and $1 + \epsilon$.
        # This keeps the KL divergence between $\pi_\theta$
        #  and $\pi_{\theta_{OLD}}$ constrained.
        # Large deviation can cause performance collapse;
        #  where the policy performance drops and doesn't recover because
        #  we are sampling from a bad policy.
        #
        # Using the normalized advantage
        #  $\bar{A_t} = \frac{\hat{A_t} - \mu(\hat{A_t})}{\sigma(\hat{A_t})}$
        #  introduces a bias to the policy gradient estimator,
        #  but it reduces variance a lot.
        clipped_ratio = ratio.clamp(min=1.0 - clip,
                                    max=1.0 + clip)
        policy_reward = torch.min(ratio * advantage,
                                  clipped_ratio * advantage)

        self.clip_fraction = (abs((ratio - 1.0)) > clip).to(torch.float).mean()

        return -policy_reward.mean()


class ClippedValueFunctionLoss(Module):
    """
    ## Clipped Value Function Loss

    \begin{align}
    V^{\pi_\theta}_{CLIP}(s_t)
     &= clip\Bigl(V^{\pi_\theta}(s_t) - \hat{V_t}, -\epsilon, +\epsilon\Bigr)
    \\
    \mathcal{L}^{VF}(\theta)
     &= \frac{1}{2} \mathbb{E} \biggl[
      max\Bigl(\bigl(V^{\pi_\theta}(s_t) - R_t\bigr)^2,
          \bigl(V^{\pi_\theta}_{CLIP}(s_t) - R_t\bigr)^2\Bigr)
     \biggr]
    \end{align}

    Clipping makes sure the value function $V_\theta$ doesn't deviate
     significantly from $V_{\theta_{OLD}}$.

    """
    def __call__(self, value: torch.Tensor, sampled_value: torch.Tensor, sampled_return: torch.Tensor, clip: float):
        clipped_value = sampled_value + (value - sampled_value).clamp(min=-clip, max=clip)
        vf_loss = torch.max((value - sampled_return) ** 2, (clipped_value - sampled_return) ** 2)
        return 0.5 * vf_loss.mean()



from typing import List

import altair as alt

try:
    import torch
except ImportError:
    torch = None


def data_to_table(series, names, step):
    table = []

    for s in range(len(series)):
        data = series[s]
        name = names[s]

        if torch is not None and isinstance(data, torch.Tensor):
            data = data.detach().cpu().numpy()

        for i in range(data.shape[0]):
            if len(data.shape) == 2:
                m = data.shape[1] // 2
                row = {'series': name, 'step': data[i, 0], 'v5': data[i, m]}
                for j in range(1, min(m, 5)):
                    row[f"v{5 - j}"] = data[i, m - j]
                    row[f"v{5 + j}"] = data[i, m + j]
            elif step is not None:
                row = {'series': name,
                       'step': step[i],
                       'v5': data[i]}
            else:
                row = {'series': name,
                       'step': i,
                       'v5': data[i]}
            table.append(row)

    return alt.Data(values=table)


def _render_density(table: alt.Data, *,
                    x_name: str,
                    levels: int,
                    alpha: float,
                    series_selection: alt.Selection = None,
                    selection: alt.Selection = None,
                    x_scale: alt.Scale = alt.Undefined,
                    y_scale: alt.Scale = alt.Undefined) -> alt.Chart:
    areas: List[alt.Chart] = []
    for i in range(1, levels):
        y = f"v{5 - i}:Q"
        y2 = f"v{5 + i}:Q"

        encode = dict(
            x=alt.X('step:Q', scale=x_scale),
            y=alt.Y(y, scale=y_scale),
            y2=alt.Y2(y2),
            color=alt.Color('series:N', scale=alt.Scale(scheme='tableau10'))
        )

        if series_selection:
            encode['opacity'] = alt.condition(series_selection, alt.value(alpha ** i),
                                              alt.value(0.001))

        areas.append(
            alt.Chart(table)
                .mark_area(opacity=alpha ** i)
                .encode(**encode)
        )

    encode = dict(x=alt.X('step:Q', scale=x_scale, title=x_name),
                  y=alt.Y("v5:Q", scale=y_scale, title='Value'),
                  color=alt.Color('series:N', scale=alt.Scale(scheme='tableau10')))

    if series_selection:
        encode['opacity'] = alt.condition(series_selection, alt.value(1), alt.value(0.001))

    line: alt.Chart = (
        alt.Chart(table)
            .mark_line()
            .encode(**encode)
    )
    if selection is not None:
        line = line.add_selection(selection)

    areas_sum = None
    for a in areas:
        if areas_sum is None:
            areas_sum = a
        else:
            areas_sum += a

    if areas_sum is not None:
        line = areas_sum + line

    if series_selection:
        line = line.add_selection(series_selection)

    return line


def render(table: alt.Data, *,
           levels=5,
           alpha=0.6,
           height: int,
           width: int,
           height_minimap: int):
    zoom = alt.selection_interval(encodings=["x", "y"])
    selection = alt.selection_multi(fields=['series'], bind='legend')

    minimaps = _render_density(table,
                               x_name='',
                               levels=levels,
                               alpha=alpha,
                               selection=zoom)

    details = _render_density(table,
                              x_name='Step',
                              levels=levels,
                              alpha=alpha,
                              series_selection=selection,
                              x_scale=alt.Scale(domain={'selection': zoom.name,
                                                        "encoding": "x"}),
                              y_scale=alt.Scale(domain={'selection': zoom.name,
                                                        "encoding": "y"}))

    minimaps = minimaps.properties(width=width, height=height_minimap)
    details = details.properties(width=width, height=height)

    return details & minimaps



from typing import Union, List, Tuple, overload, Dict

from labml.internal.util.colors import StyleCode


class Style(StyleCode):
    r"""
    Output styles
    """

    none = None
    normal = 'normal'
    bold = 'bold'
    underline = 'underline'
    light = 'light'


class Color(StyleCode):
    r"""
    Output colors
    """

    none = None
    black = 'black'
    red = 'red'
    green = 'green'
    orange = 'orange'
    blue = 'blue'
    purple = 'purple'
    cyan = 'cyan'
    white = 'white'


class Text(StyleCode):
    r"""
    Standard styles we use in labml
    """

    none = None
    danger = Color.red.value
    success = Color.green.value
    warning = Color.orange.value
    meta = Color.blue.value
    key = Color.cyan.value
    meta2 = Color.purple.value
    title = [Style.bold.value, Style.underline.value]
    heading = Style.underline.value
    value = Style.bold.value
    highlight = [Style.bold.value, Color.orange.value]
    subtle = [Style.light.value, Color.white.value]
    link = 'link'


@overload
def log():
    ...


@overload
def log(message: str, *, is_new_line: bool = True, is_reset: bool = True):
    ...


@overload
def log(message: str, color: StyleCode,
        *,
        is_new_line: bool = True, is_reset: bool = True):
    ...


@overload
def log(message: str, colors: List[StyleCode],
        *,
        is_new_line: bool = True, is_reset: bool = True):
    ...


@overload
def log(messages: List[Union[str, Tuple[str, StyleCode]]],
        *,
        is_new_line: bool = True, is_reset: bool = True):
    ...


@overload
def log(*args: Union[str, Tuple[str, StyleCode]],
        is_new_line: bool = True, is_reset: bool = True):
    ...


def log(*args, is_new_line: bool = True, is_reset: bool = True):
    r"""
    This has multiple overloads

    .. function:: log(message: str, *, is_new_line=True)
        :noindex:

    .. function:: log(message: str, color: StyleCode, *, is_new_line=True)
        :noindex:

    .. function:: log(message: str, colors: List[StyleCode], *, is_new_line=True)
        :noindex:

    .. function:: log(messages: List[Union[str, Tuple[str, StyleCode]]], *, is_new_line=True)
        :noindex:

    .. function:: log(*args: Union[str, Tuple[str, StyleCode]], is_new_line: bool = True)
        :noindex:

    Arguments:
        message (str): string to be printed
        color (StyleCode): color/style of the message
        colors (List[StyleCode]): list of colors/styles for the message
        args (Union[str, Tuple[str, StyleCode]]): list of messages
            Each element should be either a string or a tuple of string and styles.
        messages (List[Union[str, Tuple[str, StyleCode]]]): a list of messages.
            Each element should be either a string or a tuple of string and styles.

    Keyword Arguments:
        is_new_line (bool): whether to print a new line at the end

    Example::
        >>> logger.log("test")
    """
    from labml.internal.logger import logger_singleton as _internal

    if len(args) == 0:
        assert is_new_line
        _internal().log([], is_new_line=is_new_line, is_reset=is_reset)
    elif len(args) == 1:
        message = args[0]
        if isinstance(message, str):
            _internal().log([(message, None)], is_new_line=is_new_line, is_reset=is_reset)
        elif isinstance(message, list):
            _internal().log(message, is_new_line=is_new_line, is_reset=is_reset)
        else:
            raise TypeError(f'Unrecognized type: {type(message)}', message)
    elif len(args) == 2 and isinstance(args[0], str) and isinstance(args[1], StyleCode):
        _internal().log([(args[0], args[1])], is_new_line=is_new_line, is_reset=is_reset)
    else:
        _internal().log(args, is_new_line=is_new_line, is_reset=is_reset)


@overload
def inspect(items: Dict):
    ...


@overload
def inspect(items: List):
    ...


@overload
def inspect(*items: any):
    ...


@overload
def inspect(**items: any):
    ...


def inspect(*args, **kwargs):
    """
    Pretty prints a set of values.

    This has multiple overloads

    .. function:: inspect(items: Dict)
        :noindex:

    .. function:: inspect(items: List)
        :noindex:

    .. function:: inspect(*items: any)
        :noindex:
    """

    from labml.internal.logger import logger_singleton as _internal

    _internal().info(*args, **kwargs)



import sys
import time
from io import StringIO
from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from labml.internal.api import ApiCaller

WARMUP_COMMITS = 5


class ApiLogs:
    api_caller: Optional['ApiCaller']
    frequency: float

    def __init__(self):
        super().__init__()

        self.api_caller = None
        self.frequency = 1
        self.last_committed = time.time()
        self.commits_count = 0
        self.stdout = None
        self.stderr = None
        self.logger = None

    def set_api(self, api_caller: 'ApiCaller', *,
                frequency: float):
        self.api_caller = api_caller
        self.frequency = frequency
        self.check_and_flush()

    def check_and_flush(self):
        if self.api_caller is None:
            return
        if self.stdout is None and self.stderr is None and self.logger is None:
            return
        t = time.time()
        freq = self.frequency
        if self.commits_count < WARMUP_COMMITS:
            freq /= 2 ** (WARMUP_COMMITS - self.commits_count)
        if self.commits_count == 0 or t - self.last_committed > freq:
            self.last_committed = t
            self.commits_count += 1
            self.flush()

    def outputs(self, *,
                stdout_: Optional[str] = None,
                stderr_: Optional[str] = None,
                logger_: Optional[str] = None):
        if stdout_ is not None:
            if self.stdout is None:
                self.stdout = stdout_
            else:
                self.stdout += stdout_
        if stderr_ is not None:
            if self.stderr is None:
                self.stderr = stderr_
            else:
                self.stderr += stderr_
        if logger_ is not None:
            if self.logger is None:
                self.logger = logger_
            else:
                self.logger += logger_

        self.check_and_flush()

    def flush(self):
        data = {'time': time.time()}
        if self.stdout is not None:
            data['stdout'] = self.stdout
            self.stdout = None
        if self.stderr is not None:
            data['stderr'] = self.stderr
            self.stderr = None
        if self.logger is not None:
            data['logger'] = self.logger
            self.logger = None

        from labml.internal.api import Packet
        self.api_caller.push(Packet(data))


API_LOGS = ApiLogs()


class OutputStream(StringIO):
    def write(self, *args, **kwargs):  # real signature unknown
        super().write(*args, **kwargs)
        save = StringIO()
        save.write(*args, **kwargs)
        API_LOGS.outputs(**{self.type_: save.getvalue()})
        self.original.write(*args, **kwargs)

    def __init__(self, original, type_):  # real signature unknown
        super().__init__()
        self.type_ = type_
        self.original = original


def capture():
    sys.stderr = OutputStream(sys.stderr, 'stderr_')
    sys.stdout = OutputStream(sys.stdout, 'stdout_')


capture()



import pathlib
from typing import Dict
from joblib import dump, load

from . import ModelSaver, experiment_singleton


# This is under construction

class SKLearnModelSaver(ModelSaver):
    def __init__(self, name: str, model: any):
        self.name = name
        self.model = model

    def save(self, checkpoint_path: pathlib.Path) -> any:
        file_name = f"{self.name}.pth"
        dump(self.model, str(checkpoint_path / file_name))
        return file_name

    def loadl(self, checkpoint_path: pathlib.Path, info: any):
        file_name: str = info

        return load(str(checkpoint_path / file_name))


def add_models(models: Dict[str, any]):
    exp = experiment_singleton()
    savers = {name: SKLearnModelSaver(name, model) for name, model in models.items()}

    exp.checkpoint_saver.add_savers(savers)



import inspect
import warnings
from enum import Enum
from typing import List, Callable, cast, Union, Optional
from typing import TYPE_CHECKING

from .config_item import ConfigItem
from .utils import get_config_names
from ...utils.errors import ConfigsError

if TYPE_CHECKING:
    from .base import Configs


class FunctionKind(Enum):
    pass_configs = 'pass_configs'
    pass_kwargs = 'pass_kwargs'
    pass_nothing = 'pass_nothing'
    pass_params = 'pass_params'


class ConfigFunction:
    func: Callable
    kind: FunctionKind
    config_names: Union[str, List[str]]
    option_name: str
    params: List[inspect.Parameter]
    pass_params: Optional[List[ConfigItem]]

    def __get_type(self):
        key, pos = 0, 0

        for p in self.params:
            if p.kind == p.POSITIONAL_OR_KEYWORD:
                pos += 1
            elif p.kind == p.KEYWORD_ONLY:
                key += 1
            else:
                raise ConfigsError(f"Only positional or keyword only arguments should be accepted: "
                                   f"{self.config_names} - {self.option_name}")

        if self.pass_params is not None:
            if key > 0:
                raise ConfigsError(f"No keyword arguments are supported when passing configs: "
                                   f"{self.config_names} - {self.option_name}")

            if pos != len(self.pass_params):
                raise ConfigsError('Number of arguments to function should match the number of '
                                   'configs to be passed: '
                                   f"{self.config_names} - {self.option_name}")
            return FunctionKind.pass_params

        if pos == 1:
            if key != 0:
                raise ConfigsError(
                    'No keyword arguments are allowed when passing a config object: '
                    f"{self.config_names} - {self.option_name}")
            return FunctionKind.pass_configs
        elif pos == 0 and key == 0:
            return FunctionKind.pass_nothing
        else:
            warnings.warn(f"Accept configs object to your function, because it's easier to refactor, find usage etc: "
                          f"{self.config_names}: {self.option_name}."
                          f"LabML supports configs function that accept keyword arguments but this is not recommended.",
                          FutureWarning, stacklevel=5)
            if pos != 0:
                raise ConfigsError(
                    f"{self.config_names} - {self.option_name}")
            return FunctionKind.pass_kwargs

    def __get_option_name(self, option_name: str):
        if option_name is not None:
            return option_name
        else:
            return self.func.__name__

    def __get_params(self):
        func_type = type(self.func)

        if func_type == type:
            init_func = cast(object, self.func).__init__
            spec: inspect.Signature = inspect.signature(init_func)
            params: List[inspect.Parameter] = list(spec.parameters.values())
            if len(params) == 0:
                raise ConfigsError(f'Not a valid config function {self.config_names}', self.func)
            if params[0].kind != inspect.Parameter.POSITIONAL_OR_KEYWORD:
                raise ConfigsError(f'Not a valid config function {self.config_names}', self.func)
            if params[0].name != 'self':
                raise ConfigsError(f'Not a valid config function {self.config_names}', self.func)
            return params[1:]
        else:
            spec: inspect.Signature = inspect.signature(self.func)
            params: List[inspect.Parameter] = list(spec.parameters.values())
            return params

    def __init__(self, func, *,
                 config_names: Union[str, ConfigItem, List[ConfigItem], List[str]],
                 option_name: str,
                 pass_params: Optional[List[ConfigItem]] = None):
        self.func = func
        self.pass_params = pass_params
        self.config_names = get_config_names('option', self.func.__name__, config_names)
        self.option_name = self.__get_option_name(option_name)

        self.params = self.__get_params()

        self.kind = self.__get_type()

    def __call__(self, configs: 'Configs'):
        if self.kind == FunctionKind.pass_configs:
            if len(self.params) == 1:
                return self.func(configs)
            else:
                return self.func()
        elif self.kind == FunctionKind.pass_kwargs:
            kwargs = {p.name: configs.__getattribute__(p.name) for p in self.params}
            return self.func(**kwargs)
        elif self.kind == FunctionKind.pass_params:
            args = [configs.__getattribute__(p.key) for p in self.pass_params]
            return self.func(*args)
        elif self.kind == FunctionKind.pass_nothing:
            return self.func()
        else:
            assert False



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.utils.data
from torchvision import datasets, transforms

from labml import lab


class Net(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(1, 20, 5, 1)
        self.conv2 = nn.Conv2d(20, 50, 5, 1)
        self.fc1 = nn.Linear(4 * 4 * 50, 500)
        self.fc2 = nn.Linear(500, 10)

    def forward(self, x):
        x = F.relu(self.conv1(x))
        x = F.max_pool2d(x, 2, 2)
        x = F.relu(self.conv2(x))
        x = F.max_pool2d(x, 2, 2)
        x = x.view(-1, 4 * 4 * 50)
        x = F.relu(self.fc1(x))
        return self.fc2(x)


def train(epoch, model, optimizer, train_loader, device, train_log_interval):
    model.train()
    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)

        optimizer.zero_grad()
        output = model(data)
        loss = F.cross_entropy(output, target)
        loss.backward()
        optimizer.step()

        if batch_idx % train_log_interval == 0:
            print(f'train epoch: {epoch}'
                  f' [{batch_idx * len(data)}/{len(train_loader.dataset)}'
                  f' ({100. * batch_idx / len(train_loader):.0f}%)]'
                  f'\tLoss: {loss.item():.6f}')


def validate(epoch, model, valid_loader, device):
    model.eval()
    valid_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in valid_loader:
            data, target = data.to(device), target.to(device)

            output = model(data)
            valid_loss += F.cross_entropy(output, target, reduction='sum').item()
            pred = output.argmax(dim=1, keepdim=True)
            correct += pred.eq(target.view_as(pred)).sum().item()

    valid_loss /= len(valid_loader.dataset)
    valid_accuracy = 100. * correct / len(valid_loader.dataset)

    print(f'\nTest set: Average loss: {valid_loss:.4f},'
          f' Accuracy: {correct}/{len(valid_loader.dataset)}'
          f' ({valid_accuracy:.0f}%)\n')


def main():
    epochs = 10

    is_save_models = True
    train_batch_size = 64
    valid_batch_size = 1000

    use_cuda = True
    seed = 5
    train_log_interval = 10

    learning_rate = 0.01

    # get device
    is_cuda = use_cuda and torch.cuda.is_available()
    if not is_cuda:
        device = torch.device("cpu")
    else:
        device = torch.device(f"cuda:0")

    # data transform
    data_transform = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.1307,), (0.3081,))
    ])

    # train loader
    train_loader = torch.utils.data.DataLoader(
        datasets.MNIST(str(lab.get_data_path()),
                       train=True,
                       download=True,
                       transform=data_transform),
        batch_size=train_batch_size, shuffle=True)

    # valid loader
    valid_loader = torch.utils.data.DataLoader(
        datasets.MNIST(str(lab.get_data_path()),
                       train=False,
                       download=True,
                       transform=data_transform),
        batch_size=valid_batch_size, shuffle=False)

    # model
    model = Net().to(device)

    # optimizer
    optimizer = optim.Adam(model.parameters(), lr=learning_rate)

    # set seeds
    torch.manual_seed(seed)

    # training loop
    for epoch in range(1, epochs + 1):
        train(epoch, model, optimizer, train_loader, device, train_log_interval)
        validate(epoch, model, valid_loader, device)

    if is_save_models:
        torch.save(model.state_dict(), "mnist_cnn.pt")


if __name__ == '__main__':
    main()



from typing import List, Optional, NamedTuple

import numpy as np

BASIS_POINTS = [
    0,
    6.68,
    15.87,
    30.85,
    50.00,
    69.15,
    84.13,
    93.32,
    100.00
]


class Event(NamedTuple):
    step: int
    tensor: np.ndarray


class Analytics:
    def summarize(self, events):
        raise NotImplementedError()

    def summarize_scalars(self, events: List[any], points: Optional[int] = 100):
        # Shrink to 100 histograms
        if points is None:
            interval = 1
        else:
            interval = max(1, len(events) // points)

        merged = []
        results = []
        for i, e in enumerate(events):
            if i > 0 and (i + 1) % interval == 0:
                results.append(self.summarize(merged))
                merged = []
            merged.append(e)
        if len(merged) > 0:
            results.append(self.summarize(merged))

        return np.array(results)



"""
---
title: Rectified Adam (RAdam) optimizer
summary: A simple PyTorch implementation/tutorial of RAdam optimizer.
---

# Rectified Adam (RAdam) optimizer

This implementation is based on
[the official implementation](https://github.com/LiyuanLucasLiu/RAdam)
of the paper
[On the Variance of the Adaptive Learning Rate and Beyond](https://arxiv.org/abs/1908.03265).

We have implemented it as an extension to [our AMSGrad implementation](amsgrad.html)
thus requiring only the modifications to be implemented.

Adam optimizer sometimes converges to a bad local optima during the initial stages of the training;
especially when training transformers.
Researches use warmups to counter this; for the the initial training steps (warm-up stage)
they use a low learning rate.
This paper identifies the problem to be the high variance of adaptive learning rate
during initial stages of training, and counters it using a new rectification term to
reduce variance.

The paper also evaluates two variance reduction mechanisms:
* **Adam-2k**: Only compute the adaptive learning rate ($v_t$ in [Adam](adam.html)) during the first 2k steps,
without changing parameters or calculating momentum ($m_t$).
* **Adam-eps**: Adam with large $\epsilon \approx 10^{-4}$.

## Rectified Adam

Let $\sigma(g_1, ..., g_t)$ and $\psi(g_1, ..., g_t)$ be the functions to calculate
momentum and adaptive learning rate.
For Adam, they are
\begin{align}
\sigma(g_1, ..., g_t) &=  \frac{(1 - \beta_1)\sum_{i=1}^t \beta_1^{t-i} g_i}{1 - \beta_1^t} \\
\psi(g_1, ..., g_t) &=  \sqrt \frac{1 - \beta_2^t}{(1 - \beta_2)\sum_{i=1}^t \beta_2^{t-i} g_i^2}
\end{align}

### Exponential moving average as simple moving average

The distribution of exponential moving average can be approximated as a simple moving average.
\begin{align}
p\Bigg(\frac{(1-\beta_2) \sum_{i=1}^t \beta_2^{t-i} g_i^2}{1 - \beta_2^t} \Bigg) \approx
p\Bigg(\frac{\sum_{i=1}^{f(t,\beta_2)} g_{t+1-i}^2}{f(t,\beta_2)} \Bigg)
\end{align}
Here we are taking the simple moving average of the last $f(t,\beta_2)$ gradients.
$f(t,\beta_2)$ satisfies the following,
\begin{align}
\frac{(1-\beta_2) \sum_{i=1}^t \beta_2^{t-i} \cdot i}{1 - \beta_2^t} =
\frac{\sum_{i=1}^{f(t,\beta_2)} (t+1-i)}{f(t,\beta_2)}
\end{align}
which gives,
$$f(t,\beta_2) = \frac{2}{1-\beta_2} - 1 - \frac{2 t \beta_2^t}{1 - \beta_2^t}$$

### Scaled inverse chi-squared

From above we have
$$
p\Big( \psi^2(g_1, ..., g_t) \Big) \approx
p\Bigg(\frac{\sum_{i=1}^{f(t,\beta_2)} g_{t+1-i}^2}{f(t,\beta_2)} \Bigg)
$$
where $g_i \sim \mathcal{N}(0, \sigma^2)$.
Note that $sigma$ here is the standard deviation and different from $\sigma(.)$ for momentum.

[Scaled inverse chi-squared](https://en.wikipedia.org/wiki/Scaled_inverse_chi-squared_distribution)
is the distribution of squared inverse of mean of $p$ normal distributions.
$$
p\Bigg(\frac{\sum_{i=1}^{f(t,\beta_2)} g_{t+1-i}^2}{f(t,\beta_2)} \Bigg)
\sim
\text{Scale-inv} \mathcal{X}^2(\rho,\frac{1}{\sigma^2})
$$
where $\rho = f(t,\beta_2)$.

### Rectification

They prove that variance of $\psi(.)$ decreases with $\rho$ when
$\psi^2(.) \sim \text{Scale-inv} \mathcal{X}^2(\rho,\frac{1}{\sigma^2})$.

Therefore the variance is minimized at maximal $\rho$ which is
$\rho_{\infty} = \frac{2}{1-\beta_2} - 1$. Let the minimum variance be $C_{\text{var}}$

In order to ensure that the adaptive learning
rate $\psi(.)$ has consistent variance, we rectify the variance with $r$
\begin{align}
r = \sqrt{\frac{C_{\text{var}}}{Var\big[\psi(.)\big]}}
\end{align}

### Approximating $Var[\psi(.)]$

They estimate $Var[\psi(.)] \approx \frac{Var[\psi^2(.)]}{4 \mathbb{E}[\psi^2(.)}$
based on first order expansion of $\sqrt{\psi^2(.)}$
🤪 I didn't get how it was derived.

From $\text{Scale-inv} \mathcal{X}^2$ distribution we have,
\begin{align}
\mathbb{E}\big[\psi^2(.)\big] &= \frac{\rho / \sigma^2}{\rho-2} \\
Var\big[\psi^2(.)\big] &= \frac{2 \rho / \sigma^4}{(\rho-2)^2 (\rho - 2)}
\end{align}
which gives,
$$
Var[\psi(.)] \approx \frac{\rho}{2(\rho-2)(\rho-4)\sigma^2}
$$

### Rectification term

We have
\begin{align}
r &= \sqrt{\frac{C_{\text{var}}}{Var\big[\psi(.)\big]}} \\
Var[\psi(.)] &\approx \frac{\rho}{2(\rho-2)(\rho-4)\sigma^2}
\end{align}

where $C_{\text{var}}$ is $Var\big[\psi(.)\big]$ for $\rho_\infty$.
Lt $\rho$ and step $t$ be $\rho_t$, and $r_t$ be the rectification term
at step $t$.

\begin{align}
C_{\text{var}} &\approx \frac{\rho_\infty}{2(\rho_\infty-2)(\rho_\infty-4)\sigma^2} \\
Var[\psi(g_1,...,g_t)] &\approx \frac{\rho_t}{2(\rho_t-2)(\rho_t-4)\sigma^2}
\end{align}

This gives,
\begin{align}
r_t &= \sqrt{\frac{(\rho_t-2)(\rho_t-4)\rho_\infty}{(\rho_\infty-2)(\rho_\infty-4)\rho_t}}
\end{align}
"""

import math
from typing import Dict, Optional

import torch

from labml_nn.optimizers import WeightDecay
from labml_nn.optimizers.amsgrad import AMSGrad


class RAdam(AMSGrad):
    """
    ## Rectified Adam Optimizer

    This class extends from AMSAdam optimizer defined in [`amsadam.py`](amsadam.html).
    """

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8,
                 weight_decay: WeightDecay = WeightDecay(),
                 optimized_update: bool = True,
                 amsgrad=False,
                 degenerated_to_sgd=True, defaults=None):
        """
        ### Initialize the optimizer

        * `params` is the list of parameters
        * `lr` is the learning rate $\alpha$
        * `betas` is a tuple of ($\beta_1$, $\beta_2$)
        * `eps` is $\hat{\epsilon}$ or $\epsilon$ based on `optimized_update`
        * `weight_decay` is an instance of class `WeightDecay` defined in [`__init__.py`](index.html)
        * 'optimized_update' is a flag whether to optimize the bias correction of the second moment
          by doing it after adding $\epsilon$
        * `amsgrad` is a flag indicating whether to use AMSGrad or fallback to plain Adam
        * `degenerate_to_sgd` whether to use sgd when the rectification term $r_t is intractable.
        * `defaults` is a dictionary of default for group values.
         This is useful when you want to extend the class `RAdam`.
        """
        self.degenerated_to_sgd = degenerated_to_sgd
        super().__init__(params, lr, betas, eps, weight_decay, optimized_update, amsgrad, defaults)

    def step_param(self, state: Dict[str, any], group: Dict[str, any], grad: torch.Tensor, param: torch.nn.Parameter):
        """
        ### Take an update step for a given parameter tensor

        * `state` is the optimizer state of the parameter (tensor)
        * `group` stores optimizer attributes of the parameter group
        * `grad` is the current gradient tensor  $g_t$ for the parameter $\theta_{t-1}$
        * `param` is the parameter tensor $\theta_{t-1}$
        """

        # Calculate weight decay
        grad = self.weight_decay(param, grad, group)

        # Get $m_t$ and $v_t$; i.e. $\sigma(.)$ and $\psi(.)$ without bias correction
        m, v = self.get_mv(state, group, grad)

        # Calculate $t$ the number of optimizer steps
        state['step'] += 1

        # Perform *RAdam* update
        self.r_adam_update(state, group, param, m, v)

    @staticmethod
    def calc_rectification_term(beta2: float, step: int) -> Optional[float]:
        """
        ### Calculate rectification term $r_t$
        """

        # $\beta_2^t$
        beta2_t = beta2 ** step
        # $$\rho_\infty = \frac{2}{1 - \beta_2} - 1$$
        rho_inf = 2 / (1 - beta2) - 1
        # $$\rho_t = \frac{2}{1-\beta_2} - 1 - \frac{2 t \beta_2^t}{1-\beta_2^t}$$
        rho = rho_inf - 2 * step * beta2_t / (1 - beta2_t)

        # $r_t$ is tractable when $\rho_t >= 4$.
        # We are being a little more conservative since it's an approximated value
        if rho >= 5:
            # $$r_t = \sqrt{\frac{(\rho_t-2)(\rho_t-4)\rho_\infty}{(\rho_\infty-2)(\rho_\infty-4)\rho_t}}$$
            r2 = (rho - 4) / (rho_inf - 4) * (rho - 2) / rho * rho_inf / (rho_inf - 2)
            return math.sqrt(r2)
        else:
            return None

    def r_adam_update(self, state: Dict[str, any], group: Dict[str, any], param: torch.nn.Parameter,
                      m: torch.Tensor, v: torch.Tensor):
        """
        ### Do the *RAdam* parameter update

        * `state` is the optimizer state of the parameter (tensor)
        * `group` stores optimizer attributes of the parameter group
        * `param` is the parameter tensor $\theta_{t-1}$
        * `m` and `v` are the uncorrected first and second moments $m_t$ and $v_t$;
          i.e. $\sigma(.)$ and $\psi(.)$ without bias correction
        """

        # Get $\beta_1$ and $\beta_2$
        beta1, beta2 = group['betas']
        # Bias correction term for $\hat{m}_t$, $1 - \beta_1^t$
        bias_correction1 = 1 - beta1 ** state['step']
        # Bias correction term for $\hat{v}_t$, $1 - \beta_2^t$
        bias_correction2 = 1 - beta2 ** state['step']

        r = self.calc_rectification_term(beta2, state['step'])

        # If $r_t$ is intractable
        if r is not None:
            # Whether to optimize the computation by combining scalar computations
            if self.optimized_update:
                # Denominator $\sqrt{v_t} + \hat{\epsilon}$
                denominator = v.sqrt().add_(group['eps'])
                # Step size $\alpha \sqrt{r_t} * \frac{\sqrt{1-\beta_2^t}}{1-\beta_1^t}$
                step_size = self.get_lr(state, group) * math.sqrt(bias_correction2) * r / bias_correction1
                # Update parameters $\theta_t \leftarrow \theta_{t-1} - \alpha \sqrt{r_t} \frac{\sqrt{1-\beta_2^t}}{1-\beta_1^t} \cdot
                #  \frac{m_t}{\sqrt{v_t} + \hat{\epsilon}}$
                param.data.addcdiv_(m, denominator, value=-step_size)
            # Computation without optimization
            else:
                # Denominator  $\frac{\sqrt{v_t}}{\sqrt{1-\beta_2^t}} + \epsilon$
                denominator = (v.sqrt() / math.sqrt(bias_correction2)).add_(group['eps'])
                # Step size $\frac{\alpha \sqrt{r_t}}{1-\beta_1^t}$
                step_size = self.get_lr(state, group) * r / bias_correction1
                # Update parameters $\theta_t \leftarrow \theta_{t-1} - \alpha \sqrt{r_t} \cdot
                # \frac{\hat{m}_t}{\sqrt{\hat{v}_t} + \epsilon}$
                param.data.addcdiv_(m, denominator, value=-step_size)

        # If $r_t$ is intractable do a SGD with momentum
        elif self.degenerated_to_sgd:
            # Step size $\frac{\alpha}{1-\beta_1^t}$
            step_size = self.get_lr(state, group) / bias_correction1
            # Update parameters
            # $\theta_t \leftarrow \theta_{t-1} - \alpha \cdot \hat{m}_t$
            param.data.add_(m, alpha=-step_size)


def _test_rectification_term():
    """
    ### Plot $r_t$ against $t$ for various $\beta_2$

    ![Plot of r_t](radam_r_t.png)
    """
    import matplotlib.pyplot as plt
    import numpy as np

    beta2 = [0.9999, 0.999, 0.99, 0.9, 0.8, 0.6, 0.5]
    plt.plot(np.arange(1, 5_000), [[RAdam.calc_rectification_term(b, i) for b in beta2] for i in range(1, 5_000)])
    plt.legend(beta2)
    plt.title("Optimizer")
    plt.show()


if __name__ == '__main__':
    _test_rectification_term()



import math
from typing import TYPE_CHECKING, List, Tuple

from labml.internal.util.colors import StyleCode
from ...logger import Text

if TYPE_CHECKING:
    from . import Logger

try:
    import torch
except ImportError:
    torch = None

try:
    import numpy
except ImportError:
    numpy = None


def _format_int(value: int):
    return f"{value:,}"


def _format_float(value: float):
    if abs(value) < 1e-9:
        lg = 0
    else:
        lg = int(math.ceil(math.log10(abs(value)))) + 1

    decimals = 7 - lg
    decimals = max(1, decimals)
    decimals = min(6, decimals)

    fmt = "{v:8,." + str(decimals) + "f}"
    return fmt.format(v=value)


def _format_value(value: any):
    if isinstance(value, int):
        return _format_int(value)
    elif isinstance(value, float):
        return _format_float(value)
    elif numpy is not None and isinstance(value, numpy.number) and numpy.issubdtype(value, numpy.integer):
        return _format_int(int(value))
    elif numpy is not None and isinstance(value, numpy.number) and numpy.issubdtype(value, numpy.floating):
        if numpy.isnan(value):
            return 'NaN'
        return _format_float(float(value))
    elif torch is not None and isinstance(value, torch.Tensor):
        if not value.shape:
            return _format_value(value.item())
        else:
            return None
    else:
        return None


def _format_tensor(s: List[str], limit: int = 1_000, style: StyleCode = Text.value):
    res = []
    length = 0
    for p in s:
        if p in [',', ']', '[', '...']:
            res.append((p, Text.subtle))
        else:
            res.append((p, style))
            length += len(p)

        if length > limit:
            res.append((' ... ', Text.warning))
            return res, True

    return res, False


def _key_value_pair(key: any, value: any, style: StyleCode = Text.meta):
    f = _format_value(value)
    if f is None:
        f = str(value)
    return [(f'{str(key)}: ', Text.subtle),
            (f, style)]


def _render_tensor(tensor, *, new_line: str = '\n', indent: str = '', depth=0):
    if len(tensor) > 5:
        truncated = True
        idx = [0, 1, 2, '...', len(tensor) - 1]
    else:
        truncated = False
        idx = [i for i in range(len(tensor))]

    res = [indent, '[']
    if depth >= 2:
        new_line = ''
    else:
        new_line = new_line

    if new_line == '\n':
        current_indent = ' ' + indent
    else:
        indent = ''
        current_indent = ''
    if len(tensor.shape) > 1:
        res.append(new_line)
        for i in idx:
            if i == '...':
                res.append(current_indent)
                res.append('...')
            else:
                sub_res, sub_trunc = _render_tensor(tensor[i],
                                                    new_line=new_line,
                                                    indent=current_indent,
                                                    depth=depth + 1,
                                                    )
                truncated = truncated or sub_trunc
                res += sub_res
            if i != idx[-1]:
                res.append(', ')
            res.append(new_line)
    else:
        for i in idx:
            if i == '...':
                res.append('...')
            else:
                res += [_format_value(tensor[i])]
            if i != idx[-1]:
                res.append(', ')

    res.append(indent)
    res.append(']')

    return res, truncated


def _get_value_full(value: any):
    if isinstance(value, str):
        if len(value) < 500:
            return [('"', Text.subtle),
                    (value, Text.value),
                    ('" len(', Text.subtle),
                    (_format_int(len(value)), Text.meta),
                    (')', Text.subtle)]
        else:
            return [('"', Text.subtle),
                    (value[:500], Text.value),
                    (' ..." len(', Text.subtle),
                    (_format_int(len(value)), Text.meta),
                    (')', Text.subtle)]

    elif numpy is not None and isinstance(value, numpy.ndarray):
        arr, trunc = _render_tensor(value, new_line='\n')
        arr, trunc_format = _format_tensor(arr)
        if not trunc and not trunc_format:
            return arr
        else:
            return [*_key_value_pair('dtype', value.dtype),
                    '\n',
                    *_key_value_pair('shape', [s for s in value.shape], Text.value),
                    '\n',
                    *_key_value_pair('min', numpy.min(value)),
                    ' ',
                    *_key_value_pair('max', numpy.max(value)),
                    ' ',
                    *_key_value_pair('mean', numpy.mean(value)),
                    ' ',
                    *_key_value_pair('std', numpy.std(value)),
                    '\n',
                    *arr]
    elif torch is not None and isinstance(value, torch.Tensor):
        arr, trunc = _render_tensor(value, new_line='\n')
        arr, trunc_format = _format_tensor(arr)
        if not trunc and not trunc_format:
            return arr
        else:
            return [*_key_value_pair('dtype', value.dtype),
                    '\n',
                    *_key_value_pair('shape', [s for s in value.shape], Text.value),
                    '\n',
                    *_key_value_pair('min', torch.min(value).item()),
                    ' ',
                    *_key_value_pair('max', torch.max(value).item()),
                    ' ',
                    *_key_value_pair('mean', torch.mean(value.to(torch.float)).item()),
                    ' ',
                    *_key_value_pair('std', torch.std(value.to(torch.float)).item()),
                    '\n',
                    *arr]

    s = str(value)
    s = s.replace('\r', '')
    return [s]


def _shrink(s: str, style: StyleCode = Text.value, limit: int = 80):
    s = s.replace('\r', '')
    lines = s.split('\n')

    res = []
    length = 0
    for line in lines:
        if len(res) > 0:
            res.append(('\\n', Text.subtle))
        if len(line) + length < limit:
            res.append((line, style))
            length += len(line)
        else:
            res.append((line[:limit - length], style))
            res.append((' ...', Text.warning))
            break

    return res


def _get_value_line(value: any):
    f = _format_value(value)

    if f is not None:
        return [(f, Text.value)]
    elif isinstance(value, str):
        return [('"', Text.subtle)] + _shrink(value) + [('"', Text.subtle)]
    elif numpy is not None and isinstance(value, numpy.ndarray):
        return [*_format_tensor(_render_tensor(value, new_line='')[0], limit=80)[0]]
    elif torch is not None and isinstance(value, torch.Tensor):
        return [*_format_tensor(_render_tensor(value, new_line='')[0], limit=80)[0]]

    s = str(value)
    return _shrink(s)


class Inspect:
    def __init__(self, logger: 'Logger'):
        self.__logger = logger

    def _log_key_value(self, items: List[Tuple[any, any]], is_show_count=True):
        max_key_len = 0
        for k, v in items:
            max_key_len = max(max_key_len, len(str(k)))

        count = len(items)
        if count > 12:
            items = items[:10]
        for k, v in items:
            spaces = " " * (max_key_len - len(str(k)))
            self.__logger.log([(f"{spaces}{k}: ", Text.key)] +
                              _get_value_line(v))
        if count > 12:
            self.__logger.log([("...", Text.meta)])

        if is_show_count:
            self.__logger.log([
                "Total ",
                (str(count), Text.meta),
                " item(s)"])

    def info(self, *args, **kwargs):
        if len(args) == 0:
            self._log_key_value([(k, v) for k, v in kwargs.items()], False)
        elif len(args) == 1:
            assert len(kwargs.keys()) == 0
            arg = args[0]
            if isinstance(arg, list):
                self._log_key_value([(i, v) for i, v in enumerate(arg)])
            elif isinstance(arg, dict):
                keys = list(arg.keys())
                keys.sort()
                self._log_key_value([(k, arg[k]) for k in keys])
            else:
                self.__logger.log(_get_value_full(arg))
        else:
            assert len(kwargs.keys()) == 0
            self._log_key_value([(i, v) for i, v in enumerate(args)], False)



from typing import List, Optional, Tuple

import altair as alt
import numpy as np
from labml.internal.analytics.altair.utils import TABLEAU_10


def data_to_table(data: np.ndarray, x_data: np.ndarray,
                  noise: Optional[Tuple[float, float]]):
    table = []

    if noise:
        nx = np.random.rand(data.shape[0]) * noise[1]
        ny = np.random.rand(data.shape[0]) * noise[0]
    else:
        nx = np.zeros(data.shape[0])
        ny = np.zeros(data.shape[0])

    for i in range(data.shape[0]):
        if len(data.shape) == 2:  # Distribution
            m = data.shape[1] // 2
            row = {'step': x_data[i, 0],
                   'x': x_data[i, m] + nx[i],
                   'y': data[i, m] + ny[i]}
        else:
            row = {'x': x_data[i] + nx[i],
                   'y': data[i] + ny[i]}
        table.append(row)

    return alt.Data(values=table)


def _scatter_chart(table: alt.Data, *,
                   circle_size: int,
                   is_ticks: bool,
                   name: str,
                   x_name: str,
                   range_color: str,
                   height: int = None,
                   width: int = None,
                   selection: alt.Selection = None,
                   x_scale: alt.Scale = alt.Undefined,
                   y_scale: alt.Scale = alt.Undefined) -> alt.Chart:
    base = alt.Chart(table)
    if selection is not None:
        base = base.add_selection(selection)

    if not is_ticks:
        scat_x_title = x_name
        scat_y_title = name
    else:
        scat_x_title = ''
        scat_y_title = ''

    encode_kwargs = dict(
        color=alt.value(range_color)
    )
    if len(table.values) > 0 and 'step' in table.values[0]:
        encode_kwargs['opacity'] = 'step:Q'

    scat = (base
            .mark_circle(size=circle_size)
            .encode(x=alt.X('x:Q', scale=x_scale, title=scat_x_title),
                    y=alt.Y('y:Q', scale=y_scale, title=scat_y_title),
                    **encode_kwargs))

    if is_ticks:
        tick_axis = alt.Axis(labels=False, domain=False, ticks=False)

        x_ticks = base.mark_tick().encode(
            x=alt.X('x:Q', axis=tick_axis, scale=x_scale, title=x_name),
            **encode_kwargs)

        y_ticks = alt.Chart(table).mark_tick().encode(
            y=alt.X('y:Q', axis=tick_axis, scale=y_scale, title=name),
            **encode_kwargs)

        scat = scat.properties(width=width, height=height)
        x_ticks = x_ticks.properties(width=width)
        y_ticks = y_ticks.properties(height=height)
        scat = y_ticks | (scat & x_ticks)

    return scat


def render(tables: List[alt.Data], *,
           names: List[str],
           circle_size: int,
           x_name: str,
           height: int,
           width: int,
           height_minimap: int):
    zoom = alt.selection_interval(encodings=["x", "y"])

    minimaps = None
    for i, t in enumerate(tables):
        z = zoom if i == 0 else None
        minimap = _scatter_chart(t,
                                 is_ticks=False,
                                 range_color=TABLEAU_10[i % 10],
                                 name='',
                                 x_name='',
                                 selection=z,
                                 circle_size=circle_size)
        if minimaps is None:
            minimaps = minimap
        else:
            minimaps += minimap

    details = None
    for i, t in enumerate(tables):
        detail = _scatter_chart(t,
                                is_ticks=len(tables) == 1,
                                name=names[i],
                                x_name=x_name,
                                range_color=TABLEAU_10[i % 10],
                                height=height,
                                width=width,
                                x_scale=alt.Scale(domain={'selection': zoom.name,
                                                          "encoding": "x"}),
                                y_scale=alt.Scale(domain={'selection': zoom.name,
                                                          "encoding": "y"}),
                                circle_size=circle_size)

        if details is None:
            details = detail
        else:
            details += detail

    minimaps = minimaps.properties(width=width * height_minimap / height, height=height_minimap)

    return details & minimaps



from pathlib import PurePath, Path
from typing import Dict

from labml.internal.lab import lab_singleton as _internal


def get_path() -> Path:
    r"""
    Get the path to the root of the project
    """
    return _internal().path


def get_data_path() -> Path:
    r"""
    Get the path to the data folder
    """
    return _internal().data_path


def get_experiments_path() -> Path:
    r"""
    Get the path to the root of experiment logs
    """
    return _internal().experiments


def configure(configurations: Dict[str, any]):
    r"""
    Set LabML configurations through Python.
    You can set the configurations set through ``.labml.yaml``.

    `See setup guide for details. <../guide/installation_setup.html>`_
    """
    _internal().set_configurations(configurations)



import json
import socket
import threading
import time
import urllib.error
import urllib.request
from dataclasses import dataclass
from queue import Queue
from typing import Dict, Optional, Callable, Set

import labml
from labml import logger
from labml.logger import Text
from labml.utils.notice import labml_notice


@dataclass
class Packet:
    data: Dict[str, any]
    callback: Optional[Callable] = None
    idx: int = -1
    program_completed: bool = False


class _WebApiThread(threading.Thread):
    def __init__(self, url: str, state_attributes: Set[str], timeout_seconds: int):
        super().__init__(daemon=False)
        self.please_wait_count = 0
        self.timeout_seconds = timeout_seconds
        self.url = url
        self.queue: Queue[Packet] = Queue()
        self.is_stopped = False
        self.errored = False
        self.key_idx = {}
        self.data_idx = 0
        self.state_attributes = state_attributes

    def push(self, packet: Packet):
        for k in packet.data:
            self.key_idx[k] = self.data_idx
        packet.idx = self.data_idx
        self.queue.put(packet)
        self.data_idx += 1

    def stop(self):
        self.is_stopped = True
        self.push(Packet(data={}, program_completed=True))
        logger.log('Still updating LabML App, please wait for it to complete...', Text.highlight)
        self.please_wait_count = 1

    def run(self):
        while True:
            packet = self.queue.get()
            if packet.program_completed:
                logger.log()
                logger.log('Finished updating LabML App.', Text.highlight)
                return
            else:
                if self.is_stopped:
                    logger.log('Please wait' + '...' * self.please_wait_count, Text.meta,
                               is_new_line=False)
                    self.please_wait_count += 1
                retries = 0
                while True:
                    if self._process(packet):
                        break
                    else:
                        logger.log(f'Retrying again in 10 seconds ({retries})...', Text.highlight)
                        time.sleep(10)
                        retries += 1

    def _process(self, packet: Packet) -> bool:
        data = packet.data
        remove = [k for k in data if k in self.state_attributes and self.key_idx[k] != packet.idx]
        for k in remove:
            del data[k]

        if not data:
            assert packet.callback is None
            return True

        try:
            run_url = self._send(data)
        except urllib.error.HTTPError as e:
            labml_notice([f'Failed to send to {self.url}: ',
                          (str(e.code), Text.value),
                          '\n' + str(e.reason)])
            return False
        except urllib.error.URLError as e:
            labml_notice([f'Failed to connect to {self.url}\n',
                          str(e.reason)])
            return False
        except socket.timeout as e:
            labml_notice([f'{self.url} timeout\n',
                          str(e)])
            return False
        except ConnectionResetError as e:
            labml_notice([f'Connection reset by LabML App server {self.url}\n',
                          str(e)])
            return False

        if packet.callback is not None:
            packet.callback(run_url)

        return True

    def _send(self, data: Dict[str, any]):
        req = urllib.request.Request(self.url)
        req.add_header('Content-Type', 'application/json; charset=utf-8')
        data_json = json.dumps(data)
        data_json = data_json.encode('utf-8')
        req.add_header('Content-Length', str(len(data)))

        response = urllib.request.urlopen(req, data_json, timeout=self.timeout_seconds)
        content = response.read().decode('utf-8')
        result = json.loads(content)

        for e in result.get('errors', []):
            if 'error' in e:
                labml_notice(['LabML App Error: ', (e['error'] + '', Text.key), '\n',
                              (e['message'], Text.value),
                              f'App URL: {self.url}'],
                             is_danger=True)
                self.errored = True
                raise RuntimeError('LabML App Error', e)
            elif 'warning' in e:
                labml_notice(
                    ['LabML App Warning: ', (e['warning'] + ': ', Text.key), (e['message'], Text.value)])
            else:
                self.errored = True
                raise RuntimeError('Unknown error from LabML App', e)

        return result.get('url', None)


class ApiCaller:
    web_api_url: str
    params: Dict[str, str]
    thread: Optional[_WebApiThread]
    state_attributes: Set[str]

    def __init__(self, web_api_url: str, params: Dict[str, str],
                 timeout_seconds: int = 15):
        super().__init__()

        params.copy()
        params['labml_version'] = labml.__version__
        params = '&'.join([f'{k}={v}' for k, v in params.items()])

        self.web_api_url = f'{web_api_url}{params}'
        self.timeout_seconds = timeout_seconds
        self.state_attributes = set()
        self.thread = None

    def add_state_attribute(self, attr: str):
        self.state_attributes.add(attr)

    def push(self, data: Packet):
        if self.thread is None:
            self.thread = _WebApiThread(self.web_api_url, self.state_attributes, self.timeout_seconds)

        if self.thread.errored:
            raise RuntimeError('LabML App error: See above for error details')

        self.thread.push(data)
        if not self.thread.is_alive():
            self.thread.start()

    def stop(self):
        self.thread.stop()



"""
---
title: Generative Adversarial Networks experiment with MNIST
summary: This experiment generates MNIST images using multi-layer perceptron.
---

# Generative Adversarial Networks experiment with MNIST
"""

from typing import Any

import torch
import torch.nn as nn
import torch.utils.data
from torchvision import transforms

from labml import tracker, monit, experiment
from labml.configs import option
from labml_helpers.datasets.mnist import MNISTConfigs
from labml_helpers.device import DeviceConfigs
from labml_helpers.module import Module
from labml_helpers.optimizer import OptimizerConfigs
from labml_helpers.train_valid import TrainValidConfigs, hook_model_outputs, BatchIndex
from labml_nn.gan import DiscriminatorLogitsLoss, GeneratorLogitsLoss


def weights_init(m):
    classname = m.__class__.__name__
    if classname.find('Linear') != -1:
        nn.init.normal_(m.weight.data, 0.0, 0.02)
    elif classname.find('BatchNorm') != -1:
        nn.init.normal_(m.weight.data, 1.0, 0.02)
        nn.init.constant_(m.bias.data, 0)


class Generator(Module):
    """
    ### Simple MLP Generator

    This has three linear layers of increasing size with `LeakyReLU` activations.
    The final layer has a $tanh$ activation.
    """

    def __init__(self):
        super().__init__()
        layer_sizes = [256, 512, 1024]
        layers = []
        d_prev = 100
        for size in layer_sizes:
            layers = layers + [nn.Linear(d_prev, size), nn.LeakyReLU(0.2)]
            d_prev = size

        self.layers = nn.Sequential(*layers, nn.Linear(d_prev, 28 * 28), nn.Tanh())

        self.apply(weights_init)

    def forward(self, x):
        return self.layers(x).view(x.shape[0], 1, 28, 28)


class Discriminator(Module):
    """
    ### Simple MLP Discriminator

    This has three  linear layers of decreasing size with `LeakyReLU` activations.
    The final layer has a single output that gives the logit of whether input
    is real or fake. You can get the probability by calculating the sigmoid of it.
    """

    def __init__(self):
        super().__init__()
        layer_sizes = [1024, 512, 256]
        layers = []
        d_prev = 28 * 28
        for size in layer_sizes:
            layers = layers + [nn.Linear(d_prev, size), nn.LeakyReLU(0.2)]
            d_prev = size

        self.layers = nn.Sequential(*layers, nn.Linear(d_prev, 1))
        self.apply(weights_init)

    def forward(self, x):
        return self.layers(x.view(x.shape[0], -1))


class Configs(MNISTConfigs, TrainValidConfigs):
    device: torch.device = DeviceConfigs()
    epochs: int = 10

    is_save_models = True
    discriminator: Module
    generator: Module
    generator_optimizer: torch.optim.Adam
    discriminator_optimizer: torch.optim.Adam
    generator_loss: GeneratorLogitsLoss
    discriminator_loss: DiscriminatorLogitsLoss
    label_smoothing: float = 0.2
    discriminator_k: int = 1

    def init(self):
        self.state_modules = []
        self.generator = Generator().to(self.device)
        self.discriminator = Discriminator().to(self.device)
        self.generator_loss = GeneratorLogitsLoss(self.label_smoothing).to(self.device)
        self.discriminator_loss = DiscriminatorLogitsLoss(self.label_smoothing).to(self.device)

        hook_model_outputs(self.mode, self.generator, 'generator')
        hook_model_outputs(self.mode, self.discriminator, 'discriminator')
        tracker.set_scalar("loss.generator.*", True)
        tracker.set_scalar("loss.discriminator.*", True)
        tracker.set_image("generated", True, 1 / 100)

    def step(self, batch: Any, batch_idx: BatchIndex):
        self.generator.train(self.mode.is_train)
        self.discriminator.train(self.mode.is_train)

        data, target = batch[0].to(self.device), batch[1].to(self.device)

        # Increment step in training mode
        if self.mode.is_train:
            tracker.add_global_step(len(data))

        # Train the discriminator
        with monit.section("discriminator"):
            for _ in range(self.discriminator_k):
                latent = torch.randn(data.shape[0], 100, device=self.device)
                logits_true = self.discriminator(data)
                logits_false = self.discriminator(self.generator(latent).detach())
                loss_true, loss_false = self.discriminator_loss(logits_true, logits_false)
                loss = loss_true + loss_false

                # Log stuff
                tracker.add("loss.discriminator.true.", loss_true)
                tracker.add("loss.discriminator.false.", loss_false)
                tracker.add("loss.discriminator.", loss)

                # Train
                if self.mode.is_train:
                    self.discriminator_optimizer.zero_grad()
                    loss.backward()
                    if batch_idx.is_last:
                        tracker.add('discriminator', self.discriminator)
                    self.discriminator_optimizer.step()

        # Train the generator
        with monit.section("generator"):
            latent = torch.randn(data.shape[0], 100, device=self.device)
            generated_images = self.generator(latent)
            logits = self.discriminator(generated_images)
            loss = self.generator_loss(logits)

            # Log stuff
            tracker.add('generated', generated_images[0:5])
            tracker.add("loss.generator.", loss)

            # Train
            if self.mode.is_train:
                self.generator_optimizer.zero_grad()
                loss.backward()
                if batch_idx.is_last:
                    tracker.add('generator', self.generator)
                self.generator_optimizer.step()

        tracker.save()


@option(Configs.dataset_transforms)
def mnist_transforms():
    return transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.5,), (0.5,))
    ])


@option(Configs.discriminator_optimizer)
def _discriminator_optimizer(c: Configs):
    opt_conf = OptimizerConfigs()
    opt_conf.optimizer = 'Adam'
    opt_conf.parameters = c.discriminator.parameters()
    opt_conf.learning_rate = 2.5e-4
    # Setting exponent decay rate for first moment of gradient,
    # $\beta_`$ to `0.5` is important.
    # Default of `0.9` fails.
    opt_conf.betas = (0.5, 0.999)
    return opt_conf


@option(Configs.generator_optimizer)
def _generator_optimizer(c: Configs):
    opt_conf = OptimizerConfigs()
    opt_conf.optimizer = 'Adam'
    opt_conf.parameters = c.generator.parameters()
    opt_conf.learning_rate = 2.5e-4
    # Setting exponent decay rate for first moment of gradient,
    # $\beta_`$ to `0.5` is important.
    # Default of `0.9` fails.
    opt_conf.betas = (0.5, 0.999)
    return opt_conf


def main():
    conf = Configs()
    experiment.create(name='mnist_gan', comment='test')
    experiment.configs(conf,
                       {'label_smoothing': 0.01})
    with experiment.start():
        conf.run()


if __name__ == '__main__':
    main()



"""
---
title: Relative Multi-Headed Attention
summary: >
  Documented implementation with explanations of
  Relative Multi-Headed Attention from paper Transformer-XL.
---

# Relative Multi-Headed Attention

This is an implementation of 
[Transformer-XL: Attentive Language Models Beyond a Fixed-Length Context](https://arxiv.org/abs/1901.02860).
"""

import torch
from torch import nn

from labml.logger import inspect
from labml_nn.transformers.mha import MultiHeadAttention


def shift_right(x: torch.Tensor):
    """
    This method shifts $i^{th}$ row of a matrix by $i$ columns.

    If the input is `[[1, 2 ,3], [4, 5 ,6], [7, 8, 9]]`, the shifted
    result would be `[[1, 2 ,3], [0, 4, 5], [9, 0, 7]]`.
    *Ideally we should mask out the lower triangle but it's ok for our purpose*.
    """

    # Concatenate a column of zeros
    zero_pad = x.new_zeros(x.shape[0], 1, *x.shape[2:])
    x_padded = torch.cat([x, zero_pad], dim=1)

    # Remove excess elements from the end
    x_padded = x_padded.view(x.shape[1] + 1, x.shape[0], *x.shape[2:])
    x = x_padded[:-1].view_as(x)

    return x


class RelativeMultiHeadAttention(MultiHeadAttention):
    """
    ## Relative Multi-Head Attention Module

    We override [Multi-Head Attention](mha.html) module so we only need to 
    write the `get_scores` method.
    """

    def __init__(self, heads: int, d_model: int, dropout_prob: float = 0.1):
        # The linear transformations doesn't need a bias since we take care of it when
        # calculating scores.
        # However having a bias for `value` might make sense.
        super().__init__(heads, d_model, dropout_prob, False)

        self.P = 2 ** 12

        self.key_pos_embeddings = nn.Parameter(torch.zeros((self.P * 2, heads, self.d_k)), requires_grad=True)
        self.query_pos_bias = nn.Parameter(torch.zeros((heads, self.d_k)), requires_grad=True)
        self.key_pos_bias = nn.Parameter(torch.zeros((self.P * 2, heads)), requires_grad=True)

    def get_scores(self, query: torch.Tensor, key: torch.Tensor):
        """
        With absolute attention

        \begin{align}
        A^{abs}_{i,j} &= lin_q(X^q_i + P_i)^T lin_k(X^k_j + P_j) \\
                      &= Q_i^T K_j + Q_i^T U_j + V_i^T K_j + V_i^T U_j
        \end{align}

        where $Q_i$, $K_j$, $V_i$, and $U_j$ are linear transformations of
         orginal embeddings and positional encodings.

        They reason out that the attention to a given key should be the same regardless of 
        the position of query. Hence replace $V_i^T K_j$ with a constant $v^T K_j$.
        🤔 May be worthwhile testing without this assumption.

        For the second and third terms relative positional encodings are introduced.
        So $Q_i^T U_j$ is replaced with $Q_i^T R_{i - j}$ and $V_i^T U_j$ with $S_{i-j}$.

        \begin{align}
        A^{rel}_{i,j} &= Q_i^T K_j + Q_i^T R_{i - j} + v^T K_j + S_{i-j}
        \end{align}

        """

        # $R_{i-j}$ pre-shift
        key_pos_emb = self.key_pos_embeddings[self.P - query.shape[0]:self.P + key.shape[0]]
        # $S_{i-j}$ pre-shift
        key_pos_bias = self.key_pos_bias[self.P - query.shape[0]:self.P + key.shape[0]]
        # $v^T$
        query_pos_bias = self.query_pos_bias[None, None, :, :]

        # $Q_i^T K_j + v^T K_j$
        ac = torch.einsum('ibhd,jbhd->ijbh', query + query_pos_bias, key)
        # $Q_i^T R_{i - j}$ pre-shift
        b = torch.einsum('ibhd,jhd->ijbh', query, key_pos_emb)
        # $S_{i-j}$ pre-shift
        d = key_pos_bias[None, :, None, :]
        # $Q_i^T R_{i - j} + S_{i-j}$
        bd = shift_right(b + d)
        bd = bd[:, -key.shape[0]:]

        return ac + bd


def _test_shift_right():
    x = torch.tensor([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    inspect(x)
    inspect(shift_right(x))

    x = torch.arange(1, 6)[None, :, None, None].repeat(5, 1, 1, 1)
    inspect(x[:, :, 0, 0])
    inspect(shift_right(x)[:, :, 0, 0])

    x = torch.arange(1, 6)[None, :, None, None].repeat(3, 1, 1, 1)
    inspect(x[:, :, 0, 0])
    inspect(shift_right(x)[:, :, 0, 0])


if __name__ == '__main__':
    _test_shift_right()



r"""
This is an example of how you can inherit configs.
"""

import torch.nn as nn
import torch.nn.functional as F

from labml import experiment
from labml.configs import option
from labml_helpers.datasets.cifar10 import CIFAR10Configs
from labml_samples.pytorch.mnist.e_labml_helpers import Configs as MNISTExperimentConfigs


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = x.view(-1, 16 * 5 * 5)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


class Configs(CIFAR10Configs, MNISTExperimentConfigs):
    model: nn.Module = 'cifar10_model'
    dataset_name = 'CIFAR10'


@option(Configs.model)
def cifar10_model(c: Configs):
    return Net().to(c.device)


def main():
    conf = Configs()
    experiment.create(name='cifar_10')
    experiment.configs(conf,
                       {'optimizer.optimizer': 'Adam',
                        'optimizer.learning_rate': 1e-4})
    conf.set_seed.set()
    experiment.add_pytorch_models(dict(model=conf.model))
    with experiment.start():
        conf.run()


if __name__ == '__main__':
    main()



# This example is based on the following examples
# https://www.tensorflow.org/tutorials/quickstart/beginner

import tensorflow as tf

from labml import experiment
from labml.utils.keras import LabMLKerasCallback


def main():
    experiment.create(name='MNIST Keras')

    (x_train, y_train), (x_test, y_test) = tf.keras.datasets.mnist.load_data()
    x_train, x_test = x_train / 255.0, x_test / 255.0

    model = tf.keras.models.Sequential([
        tf.keras.layers.Flatten(input_shape=(28, 28)),
        tf.keras.layers.Dense(128, activation='relu'),
        tf.keras.layers.Dropout(0.2),
        tf.keras.layers.Dense(10)
    ])

    loss_fn = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True)

    model.compile(optimizer='adam',
                  loss=loss_fn,
                  metrics=['accuracy'])

    with experiment.start():
        model.fit(x_train, y_train, epochs=5, validation_data=(x_test, y_test),
                  callbacks=[LabMLKerasCallback()], verbose=None)


if __name__ == '__main__':
    main()



"""
---
title: Utilities
summary: A bunch of utility functions and classes
---

# Utilities
"""

import copy

from torch import nn

from labml_helpers.module import Module


def clone_module_list(module: Module, n: int):
    """
    ## Make a `nn.ModuleList` with clones of a given layer
    """
    return nn.ModuleList([copy.deepcopy(module) for _ in range(n)])



"""
---
title: Classify MNIST digits with Capsule Networks
summary: Code for training Capsule Networks on MNIST dataset
---

# Classify MNIST digits with Capsule Networks

This paper implements the experiment described in paper
[Dynamic Routing Between Capsules](https://arxiv.org/abs/1710.09829).
"""
from typing import Any

import torch.nn as nn
import torch.nn.functional as F
import torch.utils.data

from labml import experiment, tracker
from labml.configs import option
from labml_helpers.datasets.mnist import MNISTConfigs
from labml_helpers.metrics.accuracy import AccuracyDirect
from labml_helpers.module import Module
from labml_helpers.train_valid import SimpleTrainValidConfigs, BatchIndex
from labml_nn.capsule_networks import Squash, Router, MarginLoss


class MNISTCapsuleNetworkModel(Module):
    """
    ## Model for classifying MNIST digits
    """

    def __init__(self):
        super().__init__()
        # First convolution layer has $256$, $9 \times 9$ convolution kernels
        self.conv1 = nn.Conv2d(in_channels=1, out_channels=256, kernel_size=9, stride=1)
        # The second layer (Primary Capsules) s a convolutional capsule layer with $32$ channels
        # of convolutional $8D$ capsules ($8$ features per capsule)
        # That is, each primary capsule contains 8 convolutional units with a 9 × 9 kernel and a stride of 2.
        # In order to implement this we create a convolutional layer with $32 \times 8$ channels and
        # reshapes and permutate it's output to get the capsules of $8$ features each
        self.conv2 = nn.Conv2d(in_channels=256, out_channels=32 * 8, kernel_size=9, stride=2, padding=0)
        self.squash = Squash()

        # Routing layer gets the $32 \times 6 \times 6$ primary capsules and produces $10$ capsules.
        # Each of the primary capsules have $8$ features, while output capsules (Digit Capsules)
        # have $16$ features.
        # The routing algorithm iterates $3$ times.
        self.digit_capsules = Router(32 * 6 * 6, 10, 8, 16, 3)

        # This is the decoder mentioned in the paper.
        # It takes the outputs of the $10$ digit capsules, each with $16$ features to reproduce the
        # image. It goes through linear layers of sizes $512% and $1024$ with $ReLU$ activations.
        self.decoder = nn.Sequential(
            nn.Linear(16 * 10, 512),
            nn.ReLU(),
            nn.Linear(512, 1024),
            nn.ReLU(),
            nn.Linear(1024, 784),
            nn.Sigmoid()
        )

    def forward(self, data: torch.Tensor):
        """
        `data` are the MNIST images, with shape `[batch_size, 1, 28, 28]`
        """
        # Pass through the first convolution layer.
        # Output of this layer has shape `[batch_size, 256, 20, 20]`
        x = F.relu(self.conv1(data))
        # Pass through the second convolution layer.
        # Output of this has shape `[batch_size, 32 * 8, 6, 6]`.
        # *Note that this layer has a stride length of $2$.
        x = self.conv2(x)

        # Resize and permutate to get the capsules
        caps = x.view(x.shape[0], 8, 32 * 6 * 6).permute(0, 2, 1)
        # Squash the capsules
        caps = self.squash(caps)
        # Take them through the router to get digit capsules.
        # This has shape `[batch_size, 10, 16]`.
        caps = self.digit_capsules(caps)

        # Get masks for reconstructioon
        with torch.no_grad():
            # The prediction by the capsule network is the capsule with longest length
            pred = (caps ** 2).sum(-1).argmax(-1)
            # Create a mask to maskout all the other capsules
            mask = torch.eye(10, device=data.device)[pred]

        # Mask the digit capsules to get only the capsule that made the prediction and
        # take it through decoder to get reconstruction
        reconstructions = self.decoder((caps * mask[:, :, None]).view(x.shape[0], -1))
        # Reshape the reconstruction to match the image dimensions
        reconstructions = reconstructions.view(-1, 1, 28, 28)

        return caps, reconstructions, pred


class Configs(MNISTConfigs, SimpleTrainValidConfigs):
    """
    Configurations with MNIST data and Train & Validation setup
    """
    epochs: int = 10
    model: nn.Module = 'capsule_network_model'
    reconstruction_loss = nn.MSELoss()
    margin_loss = MarginLoss(n_labels=10)
    accuracy = AccuracyDirect()

    def init(self):
        # Print losses and accuracy to screen
        tracker.set_scalar('loss.*', True)
        tracker.set_scalar('accuracy.*', True)

        # We need to set the metrics calculate them for the epoch for training and validation
        self.state_modules = [self.accuracy]

    def step(self, batch: Any, batch_idx: BatchIndex):
        """
        This method gets called by the trainer
        """
        # Set the model mode
        self.model.train(self.mode.is_train)

        # Get the images and labels and move them to the model's device
        data, target = batch[0].to(self.device), batch[1].to(self.device)

        # Increment step in training mode
        if self.mode.is_train:
            tracker.add_global_step(len(data))

        # Whether to log activations
        with self.mode.update(is_log_activations=batch_idx.is_last):
            # Run the model
            caps, reconstructions, pred = self.model(data)

        # Calculate the total loss
        loss = self.margin_loss(caps, target) + 0.0005 * self.reconstruction_loss(reconstructions, data)
        tracker.add("loss.", loss)

        # Call accuracy metric
        self.accuracy(pred, target)

        if self.mode.is_train:
            loss.backward()

            self.optimizer.step()
            # Log parameters and gradients
            if batch_idx.is_last:
                tracker.add('model', self.model)
            self.optimizer.zero_grad()

            tracker.save()


@option(Configs.model)
def capsule_network_model(c: Configs):
    """Set the model"""
    return MNISTCapsuleNetworkModel().to(c.device)


def main():
    """
    Run the experiment
    """
    conf = Configs()
    experiment.create(name='capsule_network_mnist')
    experiment.configs(conf, {'optimizer.optimizer': 'Adam',
                              'optimizer.learning_rate': 1e-3,
                              'device.cuda_device': 1})
    with experiment.start():
        conf.run()


if __name__ == '__main__':
    main()



"""
---
title: Reinforcement Learning Algorithms
summary: >
  This is a collection of PyTorch implementations/tutorials of reinforcement learning algorithms.
  It currently includes Proximal Policy Optimization, Generalized Advantage Estimation, and
  Deep Q Networks.
---

# Reinforcement Learning Algorithms

* [Proximal Policy Optimization](ppo)
    * [This is an experiment](ppo/experiment.html) that runs a PPO agent on Atari Breakout.
    * [Generalized advantage estimation](ppo/gae.html)
* [Deep Q Networks](dqn)
    * [This is an experiment](dqn/experiment.html) that runs a DQN agent on Atari Breakout.
    * [Model](dqn/model.html) with dueling network
    * [Prioritized Experience Replay Buffer](dqn/replay_buffer.html)

[This is the implementation for OpenAI game wrapper](game.html) using `multiprocessing`.
"""


from pathlib import PurePath
from typing import Dict


class TensorboardWriter:
    def hparams(self, hparams: Dict[str, any]):
        raise NotImplemented

    def add_scalar(self, key: str, value: any, global_step: int):
        raise NotImplemented

    def add_histogram(self, key: str, value: any, global_step: int):
        raise NotImplemented

    def add_image(self, key: str, img: any, global_step: int):
        raise NotImplemented


def get_tensorboard_writer(path: PurePath) -> TensorboardWriter:
    try:
        import tensorboard
    except ImportError as e:
        raise e

    try:
        from .pytorch import PyTorchTensorboardWriter

        return PyTorchTensorboardWriter(path)
    except ImportError:
        pass

    try:
        from .tensorflow import TensorflowTensorboardWriter

        return TensorflowTensorboardWriter(path)
    except ImportError:
        raise ImportError("Tensorboard writer requires pytorch or tensorflow installed.\n"
                          "Install pytorch with pip install torch. Visit pytorch.org for details.\n"
                          "Install tensorflow with pip install tensorflow. Visit tensorflow.org for details")



"""
---
title: Multi-Headed Attention
summary: >
  This implements the Multi-Headed Attention used in transformers
  using PyTorch with explainations.
---

# Multi-Headed Attention

The implementation is inspired from [Annotated Transformer](https://nlp.seas.harvard.edu/2018/04/03/attention.html)
"""

import math
from typing import Optional

import torch
from labml import tracker
from labml_helpers.module import Module
from torch import nn as nn
from torch.nn import functional as F


class PrepareForMultiHeadAttention(Module):
    """
    This module does a linear transformation and splits the vector into given
    number of heads for multi-head attention.
    """

    def __init__(self, d_model: int, heads: int, d_k: int, bias: bool):
        super().__init__()
        self.linear = nn.Linear(d_model, heads * d_k, bias=bias)
        self.heads = heads
        self.d_k = d_k

    def __call__(self, x: torch.Tensor):
        # Input has shape `[seq_len, batch_size, d_model]`
        seq_len, batch_size, _ = x.shape

        x = self.linear(x)
        x = x.view(seq_len, batch_size, self.heads, self.d_k)

        # Output has shape `[seq_len, batch_size, heads, d_k]`
        return x


class MultiHeadAttention(Module):
    def __init__(self, heads: int, d_model: int, dropout_prob: float = 0.1, bias: bool = True):
        """
        ## Multi-Head Attention Module

        This computes multi-headed attention for given `query`, `key` and `value` vectors.
        `heads` is the number of heads.
        `d_model` is the number of features in the `query`, `key` and `value` vectors.

        $$Attention(Q, K, V) = softmax\Bigg(\frac{Q K^T}{\sqrt{d_k}}\Bigg)V$$

        """

        super().__init__()
        self.d_k = d_model // heads
        self.heads = heads

        # These transformer the `query`, `key` and `value` vectors for multi-headed attention/
        self.query = PrepareForMultiHeadAttention(d_model, heads, self.d_k, bias)
        self.key = PrepareForMultiHeadAttention(d_model, heads, self.d_k, bias)
        self.value = PrepareForMultiHeadAttention(d_model, heads, self.d_k, bias)

        # Output layer
        self.output = nn.Linear(d_model, d_model)
        self.dropout = nn.Dropout(dropout_prob)
        self.scale = 1 / math.sqrt(self.d_k)

        # We store attentions so that it can used for logging, or other computations if needed
        self.attn = None

    def get_scores(self, query: torch.Tensor, key: torch.Tensor):
        """
        ### Calculate scores between queries and keys.

        This method can be overriden for other variations like relative attention.
        """

        # Calculate $Q K^T$
        return torch.einsum('ibhd,jbhd->ijbh', query, key)

    def __call__(self, *,
                 query: torch.Tensor,
                 key: torch.Tensor,
                 value: torch.Tensor,
                 mask: Optional[torch.Tensor] = None):
        # `query`, `key` and `value`  have shape `[seq_len, batch_size, d_model]`
        seq_len, batch_size, _ = query.shape

        if mask is not None:
            # `mask` has shape `[seq_len, seq_len, batch_size]`,
            # where first dimension is the query dimension.
            # If the query dimension is equal to $1$ it will be broadcasted
            assert mask.shape[0] == 1 or mask.shape[0] == mask.shape[1]

            # Same mask applied to all heads.
            mask = mask.unsqueeze(-1)

        # Prepare `query`, `key` and `value` for attention computation
        # These will then have shape `[seq_len, batch_size, heads, d_k]`
        query = self.query(query)
        key = self.key(key)
        value = self.value(value)

        # Compute attention scores $Q K^T$
        # Results in a tensor of shape `[seq_len, seq_len, batch_size, heads]`
        scores = self.get_scores(query, key)

        # Scale scores $\frac{Q K^T}{\sqrt{d_k}}$
        scores *= self.scale

        # Apply mask
        if mask is not None:
            scores = scores.masked_fill(mask == 0, -1e9)

        # $softmax$ attention $softmax\Bigg(\frac{Q K^T}{\sqrt{d_k}}\Bigg)$
        attn = F.softmax(scores, dim=1)

        # Save attentions if debugging
        tracker.debug('attn', attn)

        # Apply dropout
        attn = self.dropout(attn)

        # Multiply by values $softmax\Bigg(\frac{Q K^T}{\sqrt{d_k}}\Bigg)V$
        x = torch.einsum("ijbh,jbhd->ibhd", attn, value)

        # Save attentions for any other calculations 
        self.attn = attn.detach()

        # Concatenate multiple heads
        x = x.reshape(seq_len, batch_size, -1)

        # Output layer
        return self.output(x)
